# sage_setup: distribution = sagemath-graphs
# cython: binding=True
# distutils: language = c++
"""
Graph coloring

This module gathers all methods related to graph coloring. Here is what it can
do :

**Proper vertex coloring**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`all_graph_colorings` | Compute all `n`-colorings a graph
    :meth:`first_coloring` | Return the first vertex coloring found
    :meth:`number_of_n_colorings` | Compute the number of `n`-colorings of a graph
    :meth:`numbers_of_colorings` | Compute the number of colorings of a graph
    :meth:`chromatic_number` | Return the chromatic number of the graph
    :meth:`vertex_coloring` | Compute vertex colorings and chromatic numbers

**Fractional relaxations**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`fractional_chromatic_number` | Return the fractional chromatic number of the graph
    :meth:`fractional_chromatic_index` | Return the fractional chromatic index of the graph

**Other colorings**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`grundy_coloring` | Compute Grundy numbers and Grundy colorings
    :meth:`b_coloring` | Compute b-chromatic numbers and b-colorings
    :meth:`edge_coloring` | Compute chromatic index and edge colorings
    :meth:`round_robin` | Compute a round-robin coloring of the complete graph on `n` vertices
    :meth:`linear_arboricity` | Compute the linear arboricity of the given graph
    :meth:`acyclic_edge_coloring` | Compute an acyclic edge coloring of the current graph


AUTHORS:

- Tom Boothby (2008-02-21): Initial version
- Carlo Hamalainen (2009-03-28): minor change: switch to C++ DLX solver
- Nathann Cohen (2009-10-24): Coloring methods using linear programming

Methods
-------
"""

# ****************************************************************************
#           Copyright (C) 2008 Tom Boothby <boothby@u.washington.edu>
#
# Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
#                         https://www.gnu.org/licenses/
# ****************************************************************************

from copy import copy
from libcpp.vector cimport vector
from libcpp.pair cimport pair

from sage.graphs.independent_sets import IndependentSets
from sage.misc.lazy_import import LazyImport

DLXCPP = LazyImport('sage.combinat.matrices.dlxcpp', 'DLXCPP')
MixedIntegerLinearProgram = LazyImport('sage.numerical.mip', 'MixedIntegerLinearProgram')


def format_coloring(data, value_only=False, hex_colors=False, vertex_color_dict=False):
    r"""
    Helper method for vertex and edge coloring methods.

    INPUT:

    - ``data`` -- either a number when ``value_only`` is ``True`` or a list of
      color classes

    - ``value_only`` -- boolean (default: ``False``); when set to ``True``, it
      simply returns ``data``

    - ``hex_colors`` -- boolean (default: ``False``); when set to ``False``,
      colors are labeled [0, 1, ..., `n - 1`], otherwise the RGB Hex labeling
      is used

    - ``vertex_color_dict`` -- boolean (default: ``False``); when set to
      ``True``, it returns a dictionary ``{vertex: color}``, otherwise it
      returns a dictionary ``{color: [list of vertices]}``

    EXAMPLES::

        sage: from sage.graphs.graph_coloring import format_coloring
        sage: color_classes = [['a', 'b'], ['c'], ['d']]
        sage: format_coloring(color_classes, value_only=True)
        [['a', 'b'], ['c'], ['d']]
        sage: format_coloring(len(color_classes), value_only=True)
        3
        sage: format_coloring(color_classes, value_only=False)
        {0: ['a', 'b'], 1: ['c'], 2: ['d']}
        sage: format_coloring(color_classes, value_only=False, hex_colors=True)         # needs sage.plot
        {'#0000ff': ['d'], '#00ff00': ['c'], '#ff0000': ['a', 'b']}
        sage: format_coloring(color_classes, value_only=False, hex_colors=False,
        ....:                 vertex_color_dict=True)
        {'a': 0, 'b': 0, 'c': 1, 'd': 2}
        sage: format_coloring(color_classes, value_only=False, hex_colors=True,         # needs sage.plot
        ....:                 vertex_color_dict=True)
        {'a': '#ff0000', 'b': '#ff0000', 'c': '#00ff00', 'd': '#0000ff'}

    TESTS::

        sage: from sage.graphs.graph_coloring import format_coloring
        sage: format_coloring([], value_only=True)
        []
        sage: format_coloring([], value_only=False, hex_colors=True)                    # needs sage.plot
        {}
        sage: format_coloring([], value_only=False, hex_colors=True,                    # needs sage.plot
        ....:                 vertex_color_dict=True)
        {}
        sage: format_coloring([], value_only=False, hex_colors=False,
        ....:                 vertex_color_dict=True)
        {}
    """
    if value_only:
        return data
    if hex_colors:
        from sage.plot.colors import rainbow
        colors = rainbow(len(data))
    else:
        colors = list(range(len(data)))
    if vertex_color_dict:
        return {u: col for col, C in zip(colors, data) for u in C}
    return {col: C for col, C in zip(colors, data) if C}


def all_graph_colorings(G, n, count_only=False, hex_colors=False,
                        vertex_color_dict=False, color_classes=False):
    r"""
    Compute all `n`-colorings of a graph.

    This method casts the graph coloring problem into an exact cover problem,
    and passes this into an implementation of the Dancing Links algorithm
    described by Knuth (who attributes the idea to Hitotumatu and Noshita).

    INPUT:

    - ``G`` -- a graph

    - ``n`` -- positive integer; the number of colors

    - ``count_only`` -- boolean (default: ``False``); when set to ``True``, it
      returns 1 for each coloring and ignores other parameters

    - ``hex_colors`` -- boolean (default: ``False``); when set to ``False``,
      colors are labeled [0, 1, ..., `n - 1`], otherwise the RGB Hex labeling
      is used

    - ``vertex_color_dict`` -- boolean (default: ``False``); when set to
      ``True``, it returns a dictionary ``{vertex: color}``, otherwise it
      returns a dictionary ``{color: [list of vertices]}``

    - ``color_classes`` -- boolean (default: ``False``); when set to ``True``,
      the method returns only a list of the color classes and ignores parameters
      ``hex_colors`` and ``vertex_color_dict``

    .. WARNING::

        This method considers only colorings using exactly `n` colors, even if a
        coloring using fewer colors can be found.

    The construction works as follows. Columns:

    * The first `|V|` columns correspond to a vertex -- a `1` in this column
      indicates that this vertex has a color.

    * After those `|V|` columns, we add `n*|E|` columns -- a `1` in these
      columns indicate that a particular edge is incident to a vertex with a
      certain color.

    Rows:

    * For each vertex, add `n` rows; one for each color `c`. Place a `1` in the
      column corresponding to the vertex, and a `1` in the appropriate column
      for each edge incident to the vertex, indicating that that edge is
      incident to the color `c`.

    * If `n > 2`, the above construction cannot be exactly covered since each
      edge will be incident to only two vertices (and hence two colors) - so we
      add `n*|E|` rows, each one containing a `1` for each of the `n*|E|`
      columns. These get added to the cover solutions "for free" during the
      backtracking.

    Note that this construction results in `n*|V| + 2*n*|E| + n*|E|` entries in
    the matrix.  The Dancing Links algorithm uses a sparse representation, so if
    the graph is simple, `|E| \leq |V|^2` and `n <= |V|`, this construction runs
    in `O(|V|^3)` time.  Back-conversion to a coloring solution is a simple scan
    of the solutions, which will contain `|V| + (n-2)*|E|` entries, so runs in
    `O(|V|^3)` time also. For most graphs, the conversion will be much faster
    -- for example, a planar graph will be transformed for `4`-coloring in
    linear time since `|E| = O(|V|)`.

    REFERENCES:

    http://www-cs-staff.stanford.edu/~uno/papers/dancing-color.ps.gz

    EXAMPLES::

        sage: from sage.graphs.graph_coloring import all_graph_colorings
        sage: G = Graph({0: [1, 2, 3], 1: [2]})
        sage: n = 0
        sage: for C in all_graph_colorings(G, 3, hex_colors=True):                      # needs sage.plot
        ....:     parts = [C[k] for k in C]
        ....:     for P in parts:
        ....:         l = len(P)
        ....:         for i in range(l):
        ....:             for j in range(i + 1, l):
        ....:                 if G.has_edge(P[i], P[j]):
        ....:                     raise RuntimeError("Coloring Failed.")
        ....:     n += 1
        sage: print("G has %s 3-colorings." % n)                                        # needs sage.plot
        G has 12 3-colorings.

    TESTS::

        sage: G = Graph({0: [1, 2, 3], 1: [2]})
        sage: for C in all_graph_colorings(G, 0):
        ....:     print(C)
        sage: for C in all_graph_colorings(G, -1):
        ....:     print(C)
        Traceback (most recent call last):
        ...
        ValueError: n must be nonnegative
        sage: G = Graph({0: [1], 1: [2]})
        sage: for c in all_graph_colorings(G, 2, vertex_color_dict=True):
        ....:     print(c)
        {0: 0, 2: 0, 1: 1}
        {1: 0, 0: 1, 2: 1}
        sage: for c in all_graph_colorings(G, 2, hex_colors=True):                      # needs sage.plot
        ....:     print(sorted(c.items()))
        [('#00ffff', [1]), ('#ff0000', [0, 2])]
        [('#00ffff', [0, 2]), ('#ff0000', [1])]
        sage: for c in all_graph_colorings(G, 2, hex_colors=True,                       # needs sage.plot
        ....:                              vertex_color_dict=True):
        ....:     print(c)
        {0: '#ff0000', 2: '#ff0000', 1: '#00ffff'}
        {1: '#ff0000', 0: '#00ffff', 2: '#00ffff'}
        sage: for c in all_graph_colorings(G, 2, vertex_color_dict=True):
        ....:     print(c)
        {0: 0, 2: 0, 1: 1}
        {1: 0, 0: 1, 2: 1}
        sage: for c in all_graph_colorings(G, 2, count_only=True, vertex_color_dict=True):
        ....:     print(c)
        1
        1
        sage: for c in all_graph_colorings(G, 2, color_classes=True):
        ....:     print(c)
        [[0, 2], [1]]
        [[1], [0, 2]]
    """
    G._scream_if_not_simple(allow_multiple_edges=True)

    if not n or n > G.order():
        return
    if n < 0:
        raise ValueError("n must be nonnegative")

    cdef list V = list(G)

    cdef int nV = G.order()
    cdef int nE = G.size()

    cdef vector[pair[int, vector[int]]] ones
    cdef dict Vd = {}
    cdef dict colormap = {}
    cdef int k = 0
    for i, v in enumerate(V):
        Vd[v] = i
        for c in range(n):
            ones.push_back((k, [i]))
            colormap[k] = (v, c)
            k += 1

    cdef int kk = nV
    cdef int v0, v1
    for e in G.edges(labels=False, sort=False):
        v0 = n * Vd[e[0]]
        v1 = n * Vd[e[1]]
        for c in range(n):
            ones[v0].second.push_back(kk + c)
            ones[v1].second.push_back(kk + c)
            v0 += 1
            v1 += 1
        kk += n

    if n > 2:
        for i in range(n * nE):
            ones.push_back((k + i, [nV + i]))

    cdef list ones_second = [ones[i].second for i in range(len(ones))]
    cdef list coloring
    cdef set used_colors

    try:
        for a in DLXCPP(ones_second):
            coloring = [[] for _ in range(n)]
            used_colors = set()
            if count_only:
                used_colors = set(colormap[x][1] for x in a if x in colormap)
            else:
                for x in a:
                    if x in colormap:
                        v, c = colormap[x]
                        used_colors.add(c)
                        coloring[c].append(v)
            if len(used_colors) == n:
                if count_only:
                    yield 1
                else:
                    yield format_coloring(coloring, value_only=color_classes,
                                          hex_colors=hex_colors,
                                          vertex_color_dict=vertex_color_dict)
    except RuntimeError:
        raise RuntimeError("too much recursion, Graph coloring failed")


cpdef first_coloring(G, n=0, hex_colors=False):
    r"""
    Return the first vertex coloring found.

    If a natural number `n` is provided, returns the first found coloring with
    at least `n` colors. That is, `n` is a lower bound on the number of colors
    to use.

    INPUT:

    - ``n`` -- integer (default: 0); the minimal number of colors to try

    - ``hex_colors`` -- boolean (default: ``False``); when set to ``True``, the
      partition returned is a dictionary whose keys are colors and whose values
      are the color classes (ideal for plotting)

    EXAMPLES::

        sage: from sage.graphs.graph_coloring import first_coloring
        sage: G = Graph({0: [1, 2, 3], 1: [2]})
        sage: sorted(first_coloring(G, 3))
        [[0], [1, 3], [2]]

    TESTS:

    :issue:`33554` is fixed::

        sage: P3 = graphs.PathGraph(3)
        sage: [len(graph_coloring.first_coloring(P3, k)) for k in range(P3.order() + 1)]
        [2, 2, 2, 3]
    """
    G._scream_if_not_simple(allow_multiple_edges=True)
    cdef int o = G.order()
    for m in range(n, o + 1):
        for C in all_graph_colorings(G, m, hex_colors=hex_colors, color_classes=not hex_colors):
            return C


cpdef number_of_n_colorings(G, n):
    r"""
    Compute the number of `n`-colorings of a graph.

    INPUT:

    - ``G`` -- a graph

    - ``n`` -- positive integer; the number of colors

    EXAMPLES::

        sage: from sage.graphs.graph_coloring import number_of_n_colorings
        sage: G = Graph({0: [1, 2, 3], 1: [2]})
        sage: number_of_n_colorings(G, 3)
        12
    """
    # Take care of the stupid stuff
    if not n:
        return int(G.order() == 0)
    if n == 1:
        return int(G.size() == 0)
    if n < 0:
        # negative colors?? what does that even mean?
        return 0

    m = 0
    for C in all_graph_colorings(G, n, count_only=True):
        m += 1
    return m


cpdef numbers_of_colorings(G):
    r"""
    Compute the number of colorings of a graph.

    Return the number of `n`-colorings of the graph `G` for all `n` from `0` to
    `|V|`.

    EXAMPLES::

        sage: from sage.graphs.graph_coloring import numbers_of_colorings
        sage: G = Graph({0: [1, 2, 3], 1: [2]})
        sage: numbers_of_colorings(G)
        [0, 0, 0, 12, 24]
    """
    cdef int o = G.order()
    cdef list answer = [number_of_n_colorings(G, n) for n in range(o + 1)]
    return answer


cpdef chromatic_number(G):
    r"""
    Return the chromatic number of the graph.

    The chromatic number is the minimal number of colors needed to color the
    vertices of the graph `G`.

    EXAMPLES::

        sage: from sage.graphs.graph_coloring import chromatic_number
        sage: G = Graph({0: [1, 2, 3], 1: [2]})
        sage: chromatic_number(G)                               # needs cliquer
        3

        sage: G = graphs.PetersenGraph()
        sage: G.chromatic_number()                              # needs cliquer
        3
    """
    G._scream_if_not_simple(allow_multiple_edges=True)
    cdef int o = G.order()
    if not o:
        return 0
    if not G.size():
        return 1
    elif G.is_bipartite():  # can we do it in linear time?
        return 2

    # counting cliques is faster than our brute-force method...
    cdef int m = G.clique_number()
    if m >= o - 1:
        # marginal improvement... if there is an o-1 clique and not an o clique,
        # don't waste our time coloring.
        return m
    for n in range(m, o + 1):
        for C in all_graph_colorings(G, n, count_only=True):
            return n


def vertex_coloring(g, k=None, value_only=False, hex_colors=False, solver=None, verbose=0,
                    *, integrality_tolerance=1e-3):
    r"""
    Compute Vertex colorings and chromatic numbers.

    This function can compute the chromatic number of the given graph or test
    its `k`-colorability.

    See the :wikipedia:`Graph_coloring` for further details on graph coloring.

    INPUT:

    - ``g`` -- a graph

    - ``k`` -- integer (default: ``None``); tests whether the graph is
      `k`-colorable.  The function returns a partition of the vertex set in `k`
      independent sets if possible and ``False`` otherwise.

    - ``value_only`` -- boolean (default: ``False``):

      - When set to ``True``, only the chromatic number is returned.

      - When set to ``False`` (default), a partition of the vertex set into
        independent sets is returned if possible.

    - ``hex_colors`` -- boolean (default: ``False``); when set to ``True``, the
      partition returned is a dictionary whose keys are colors and whose values
      are the color classes (ideal for plotting).

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    OUTPUT:

    - If ``k=None`` and ``value_only=False``, then return a partition of the
      vertex set into the minimum possible of independent sets.

    - If ``k=None`` and ``value_only=True``, return the chromatic number.

    - If ``k`` is set and ``value_only=None``, return ``False`` if the graph is
      not `k`-colorable, and a partition of the vertex set into `k` independent
      sets otherwise.

    - If ``k`` is set and ``value_only=True``, test whether the graph is
      `k`-colorable, and return ``True`` or ``False`` accordingly.

    EXAMPLES::

       sage: from sage.graphs.graph_coloring import vertex_coloring
       sage: g = graphs.PetersenGraph()
       sage: vertex_coloring(g, value_only=True)                # needs cliquer sage.numerical.mip
       3

    TESTS:

    Empty graph::

       sage: from sage.graphs.graph_coloring import vertex_coloring
       sage: empty = Graph()
       sage: vertex_coloring(empty, value_only=True)
       0
       sage: vertex_coloring(empty, hex_colors=True)
       {}
       sage: vertex_coloring(empty)
       []

    :issue:`33559` is fixed::

        sage: G = Graph('MgCgS?_O@IeTHKG??')
        sage: len(G.coloring(algorithm='MILP'))                                         # needs cliquer sage.numerical.mip
        4
    """
    g._scream_if_not_simple(allow_multiple_edges=True)
    cdef list colorings
    cdef set vertices
    cdef list deg
    cdef list neighbors
    cdef list classes

    # If k is None, tries to find an optimal coloring
    if k is None:
        # No need to start a linear program if the graph is an
        # independent set, is bipartite, or is empty.
        # - Empty graph
        if not g.order():
            if value_only:
                return 0
            elif hex_colors:
                return dict()
            return []
        # - Independent set
        if not g.size():
            if value_only:
                return 1
            return format_coloring([list(g)], value_only=not hex_colors,
                                   hex_colors=hex_colors)
        # - Bipartite set
        if g.is_bipartite():
            if value_only:
                return 2
            return format_coloring(g.bipartite_sets(),
                                   value_only=not hex_colors,
                                   hex_colors=hex_colors)

        # - No need to try any k smaller than the maximum clique in the graph
        # - No need to try k less than |G|/alpha(G), as each color
        #   class is at most alpha(G)
        # - max, because we know it is not bipartite
        from math import ceil
        k = int(max([3, g.clique_number(), ceil(g.order() / len(g.independent_set()))]))

        while True:
            # tries to color the graph, increasing k each time it fails.
            tmp = vertex_coloring(g, k=k, value_only=value_only,
                                  hex_colors=hex_colors, solver=solver, verbose=verbose,
                                  integrality_tolerance=integrality_tolerance)
            if tmp is not False:
                if value_only:
                    return k
                else:
                    return tmp
            k += 1
    else:
        # Is the graph empty?
        # If the graph is empty, something should be returned.
        # This is not so stupid, as the graph could be emptied
        # by the test of degeneracy.
        if not g.order():
            if value_only:
                return True
            return format_coloring([[] for i in range(k)],
                                   value_only=not hex_colors,
                                   hex_colors=hex_colors)
        # Is the graph connected?
        # This is not so stupid, as the graph could be disconnected
        # by the test of degeneracy (as previously).
        if not g.is_connected():
            if value_only:
                for component in g.connected_components(sort=False):
                    tmp = vertex_coloring(g.subgraph(component), k=k,
                                          value_only=value_only,
                                          hex_colors=hex_colors,
                                          solver=solver, verbose=verbose,
                                          integrality_tolerance=integrality_tolerance)
                    if tmp is False:
                        return False
                return True
            colorings = []
            for component in g.connected_components(sort=False):
                tmp = vertex_coloring(g.subgraph(component), k=k,
                                      value_only=value_only,
                                      hex_colors=False,
                                      solver=solver, verbose=verbose,
                                      integrality_tolerance=integrality_tolerance)
                if tmp is False:
                    return False
                colorings.append(tmp)
            value = [[] for color in range(k)]
            for color in range(k):
                for component in colorings:
                    value[color].extend(component[color])

            return format_coloring(value, value_only=not hex_colors,
                                   hex_colors=hex_colors)

        # Degeneracy
        # Vertices whose degree is less than k are of no importance in
        # the coloring.
        if min(g.degree()) < k:
            vertices = set(g)
            deg = []
            tmp = [v for v in vertices if g.degree(v) < k]
            while tmp:
                v = tmp.pop(0)
                neighbors = list(set(g.neighbor_iterator(v)) & vertices)
                if v in vertices and len(neighbors) < k:
                    vertices.remove(v)
                    tmp.extend(neighbors)
                    deg.append(v)
            if value_only:
                return vertex_coloring(g.subgraph(list(vertices)), k=k,
                                       value_only=value_only,
                                       hex_colors=hex_colors,
                                       solver=solver, verbose=verbose,
                                       integrality_tolerance=integrality_tolerance)
            value = vertex_coloring(g.subgraph(list(vertices)), k=k,
                                    value_only=value_only,
                                    hex_colors=False,
                                    solver=solver, verbose=verbose,
                                    integrality_tolerance=integrality_tolerance)
            if value is False:
                return False
            while deg:
                for classe in value:
                    if set(classe).isdisjoint(g.neighbor_iterator(deg[-1])):
                        classe.append(deg[-1])
                        deg.pop(-1)
                        break

            return format_coloring(value, value_only=not hex_colors,
                                   hex_colors=hex_colors)

        p = MixedIntegerLinearProgram(maximization=True, solver=solver)
        color = p.new_variable(binary=True)

        # a vertex has exactly one color
        for v in g:
            p.add_constraint(p.sum(color[v, i] for i in range(k)), min=1, max=1)

        # adjacent vertices have different colors
        for u, v in g.edge_iterator(labels=None):
            for i in range(k):
                p.add_constraint(color[u, i] + color[v, i], max=1)

        # The first vertex is colored with 1. It costs nothing to say
        # it, and it can help.
        p.add_constraint(color[next(g.vertex_iterator()), 0],  max=1, min=1)

        from sage.numerical.mip import MIPSolverException
        try:
            if value_only:
                p.solve(objective_only=True, log=verbose)
                return True
            else:
                p.solve(log=verbose)
        except MIPSolverException:
            return False

        color = p.get_values(color, convert=bool, tolerance=integrality_tolerance)
        # builds the color classes
        classes = [[] for i in range(k)]

        for v in g:
            for i in range(k):
                if color[v, i]:
                    classes[i].append(v)
                    break

        return format_coloring(classes, value_only=not hex_colors,
                               hex_colors=hex_colors)


# Fractional relaxations
def fractional_chromatic_number(G, solver='PPL', verbose=0,
                                check_components=True, check_bipartite=True):
    r"""
    Return the fractional chromatic number of the graph.

    Fractional coloring is a relaxed version of vertex coloring with several
    equivalent definitions, such as the optimum value in a linear relaxation of
    the integer program that gives the usual chromatic number. It is also equal
    to the fractional clique number by LP-duality.

    ALGORITHM:

    The fractional chromatic number is computed via the usual Linear Program.
    The LP solved by sage is essentially,

    .. MATH::

        \mbox{Minimize : }&\sum_{I\in \mathcal{I}(G)} x_{I}\\
        \mbox{Such that : }&\\
        &\forall v\in V(G), \sum_{I\in \mathcal{I}(G),\, v\in I}x_{v}\geq 1\\
        &\forall I\in \mathcal{I}(G), x_{I} \geq 0

    where `\mathcal{I}(G)` is the set of maximal independent sets of `G` (see
    Section 2.1 of [CFKPR2010]_ to know why it is sufficient to consider maximal
    independent sets). As optional optimisations, we construct the LP on each
    biconnected component of `G` (and output the maximum value), and avoid using
    the LP if G is bipartite (as then the output must be 1 or 2).

    .. NOTE::

        Computing the fractional chromatic number can be very slow. Since the
        variables of the LP are independent sets, in general the LP has size
        exponential in the order of the graph. In the current implementation a
        list of all maximal independent sets is created and stored, which can be
        both slow and memory-hungry.

    INPUT:

    - ``G`` -- a graph

    - ``solver`` -- (default: ``'PPL'``) specify a Linear Program (LP) solver
      to be used. If set to ``None``, the default one is used. For more
      information on LP solvers and which default solver is used, see the method
      :meth:`solve <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the
      class :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

      .. NOTE::

          The default solver used here is ``'PPL'`` which provides exact
          results, i.e. a rational number, although this may be slower that
          using other solvers.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity of
      the LP solver

    - ``check_components`` -- boolean (default: ``True``); whether the method is
      called on each biconnected component of `G`

    - ``check_bipartite`` -- boolean (default: ``True``); whether the graph is
      checked for bipartiteness. If the graph is bipartite then we can avoid
      creating and solving the LP.

    EXAMPLES:

    The fractional chromatic number of a `C_5` is `5/2`::

        sage: g = graphs.CycleGraph(5)
        sage: g.fractional_chromatic_number()                                           # needs sage.numerical.mip
        5/2

    TESTS::

        sage: G = graphs.RandomGNP(20, .2)
        sage: a = G.fractional_chromatic_number(check_components=True)                  # needs sage.numerical.mip
        sage: b = G.fractional_chromatic_number(check_components=False)                 # needs sage.numerical.mip
        sage: a == b                                                                    # needs sage.numerical.mip
        True
    """
    G._scream_if_not_simple()

    if not G.order():
        return 0
    if not G.size():
        # The fractional chromatic number of an independent set is 1
        return 1
    if check_bipartite and G.is_bipartite():
        # at this point we've already ascertained g.size() > 0
        # so g is a bipartite graph with at least one edge
        return 2
    if check_components:
        return max(fractional_chromatic_number(G.subgraph(b), solver=solver,
                                               verbose=verbose,
                                               check_components=False,
                                               check_bipartite=check_bipartite)
                   for b in G.blocks_and_cut_vertices()[0])

    Is = [frozenset(S) for S in IndependentSets(G, maximal=True)]

    # Initialize LP for fractional chromatic number, we want to minimize the
    # total weight
    p = MixedIntegerLinearProgram(solver=solver, maximization=False)

    # One nonnegative variable per maximal independent set
    w = p.new_variable(nonnegative=True)

    # the objective is the sum of weights of the independent sets
    p.set_objective(p.sum(w[S] for S in Is))

    # such that each vertex gets total weight at least 1
    for v in G:
        p.add_constraint(p.sum(w[S] for S in Is if v in S), min=1)

    obj = p.solve(log=verbose)

    return obj


def fractional_chromatic_index(G, solver='PPL', verbose_constraints=False, verbose=0):
    r"""
    Return the fractional chromatic index of the graph.

    The fractional chromatic index is a relaxed version of edge-coloring. An
    edge coloring of a graph being actually a covering of its edges into the
    smallest possible number of matchings, the fractional chromatic index of a
    graph `G` is the smallest real value `\chi_f(G)` such that there exists a
    list of matchings `M_1, \ldots, M_k` of `G` and coefficients `\alpha_1,
    \ldots, \alpha_k` with the property that each edge is covered by the
    matchings in the following relaxed way

    .. MATH::

        \forall e \in E(G), \sum_{e \in M_i} \alpha_i \geq 1.

    For more information, see the :wikipedia:`Fractional_coloring`.

    ALGORITHM:

    The fractional chromatic index is computed through Linear Programming
    through its dual. The LP solved by sage is actually:

    .. MATH::

        \mbox{Maximize : }&\sum_{e\in E(G)} r_{e}\\
        \mbox{Such that : }&\\
        &\forall M\text{ matching }\subseteq G, \sum_{e\in M}r_{v}\leq 1\\

    INPUT:

    - ``G`` -- a graph

    - ``solver`` -- (default: ``'PPL'``) specify a Linear Program (LP) solver
      to be used. If set to ``None``, the default one is used. For more
      information on LP solvers and which default solver is used, see the method
      :meth:`solve <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the
      class :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

      .. NOTE::

          The default solver used here is ``'PPL'`` which provides exact
          results, i.e. a rational number, although this may be slower that
          using other solvers. Be aware that this method may loop endlessly when
          using some non exact solvers as reported in :issue:`23658` and
          :issue:`23798`.

    - ``verbose_constraints`` -- boolean (default: ``False``); whether to
      display which constraints are being generated

    - ``verbose`` -- integer (default: 0); sets the level of verbosity of the
      LP solver

    EXAMPLES:

    The fractional chromatic index of a `C_5` is `5/2`::

        sage: g = graphs.CycleGraph(5)
        sage: g.fractional_chromatic_index()                                            # needs sage.numerical.mip
        5/2

    TESTS:

    Issue reported in :issue:`23658` and :issue:`23798` with non exact
    solvers::

        sage: g = graphs.PetersenGraph()
        sage: g.fractional_chromatic_index(solver='GLPK')       # known bug             # needs sage.numerical.mip
        3.0
        sage: g.fractional_chromatic_index(solver='PPL')                                # needs sage.numerical.mip
        3
    """
    G._scream_if_not_simple()

    if not G.order():
        return 0
    if not G.size():
        return 1

    frozen_edges = [frozenset(e) for e in G.edges(labels=False, sort=False)]

    # Initialize LP for maximum weight matching
    M = MixedIntegerLinearProgram(solver=solver, constraint_generation=True)

    # One variable per edge
    b = M.new_variable(binary=True, nonnegative=True)

    # We want to select at most one incident edge per vertex (matching)
    for u in G:
        M.add_constraint(M.sum(b[frozenset(e)] for e in G.edges_incident(u, labels=False)) <= 1)

    #
    # Initialize LP for fractional chromatic number
    p = MixedIntegerLinearProgram(solver=solver, constraint_generation=True)

    # One variable per edge
    r = p.new_variable(nonnegative=True)

    # We want to maximize the sum of weights on the edges
    p.set_objective(p.sum(r[fe] for fe in frozen_edges))

    # Each edge being by itself a matching, its weight cannot be more than 1
    for fe in frozen_edges:
        p.add_constraint(r[fe] <= 1)

    obj = p.solve(log=verbose)

    while True:

        # Update the weights of edges for the matching problem
        M.set_objective(M.sum(p.get_values(r[fe]) * b[fe] for fe in frozen_edges))

        # If the maximum matching has weight at most 1, we are done !
        if M.solve(log=verbose) <= 1:
            break

        # Otherwise, we add a new constraint
        matching = [fe for fe in frozen_edges if M.get_values(b[fe]) == 1]
        p.add_constraint(p.sum(r[fe] for fe in matching) <= 1)
        if verbose_constraints:
            print("Adding a constraint on matching : {}".format(matching))

        # And solve again
        obj = p.solve(log=verbose)

    # Accomplished !
    return obj


def grundy_coloring(g, k, value_only=True, solver=None, verbose=0,
                    *, integrality_tolerance=1e-3):
    r"""
    Compute Grundy numbers and Grundy colorings.

    The method computes the worst-case of a first-fit coloring with less than
    `k` colors.

    Definition:

    A first-fit coloring is obtained by sequentially coloring the vertices of a
    graph, assigning them the smallest color not already assigned to one of its
    neighbors. The result is clearly a proper coloring, which usually requires
    much more colors than an optimal vertex coloring of the graph, and heavily
    depends on the ordering of the vertices.

    The number of colors required by the worst-case application of this
    algorithm on a graph `G` is called the Grundy number, written `\Gamma (G)`.

    Equivalent formulation:

    Equivalently, a Grundy coloring is a proper vertex coloring such that any
    vertex colored with `i` has, for every `j < i`, a neighbor colored with
    `j`. This can define a Linear Program, which is used here to compute the
    Grundy number of a graph.

    .. NOTE::

       This method computes a grundy coloring using at *MOST* `k` colors. If
       this method returns a value equal to `k`, it cannot be assumed that `k`
       is equal to `\Gamma(G)`. Meanwhile, if it returns any value `k' < k`,
       this is a certificate that the Grundy number of the given graph is `k'`.

       As `\Gamma(G)\leq \Delta(G)+1`, it can also be assumed that `\Gamma(G) =
       k` if ``grundy_coloring(g, k)`` returns `k` when `k = \Delta(G) +1`.

    INPUT:

    - ``k`` -- integer;  maximum number of colors

    - ``value_only`` -- boolean (default: ``True``); when set to ``True``, only
      the number of colors is returned. Otherwise, the pair ``(nb_colors,
      coloring)`` is returned, where ``coloring`` is a dictionary associating
      its color (integer) to each vertex of the graph.

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    ALGORITHM:

    Integer Linear Program.

    EXAMPLES:

    The Grundy number of a `P_4` is equal to 3::

        sage: from sage.graphs.graph_coloring import grundy_coloring
        sage: g = graphs.PathGraph(4)
        sage: grundy_coloring(g, 4)                                                     # needs sage.numerical.mip
        3

    The Grundy number of the PetersenGraph is equal to 4::

        sage: g = graphs.PetersenGraph()
        sage: grundy_coloring(g, 5)                                                     # needs sage.numerical.mip
        4

    It would have been sufficient to set the value of ``k`` to 4 in
    this case, as `4 = \Delta(G)+1`.
    """
    g._scream_if_not_simple(allow_multiple_edges=True)

    p = MixedIntegerLinearProgram(solver=solver)

    # b[v,i] is set to 1 if and only if v is colored with i
    b = p.new_variable(binary=True)

    # is_used[i] is set to 1 if and only if color[i] is used by some vertex
    is_used = p.new_variable(binary=True)

    # Each vertex is in exactly one color class
    for v in g:
        p.add_constraint(p.sum(b[v, i] for i in range(k)), max=1, min=1)

    # Two adjacent vertices have different colors
    for u, v in g.edge_iterator(labels=None):
        for i in range(k):
            p.add_constraint(b[v, i] + b[u, i], max=1)

    # The following constraints ensure that if v is colored with i, then it has
    # a neighbor colored with j for every j<i

    for i in range(k):
        for j in range(i):
            for v in g:
                # If b[v,i] == 0, then the following constraint is always
                # satisfied, as a sum of binary variables is always positive.
                # If it is equal to 1, then at least one of the other variables
                # must be set to 1 too.

                p.add_constraint(p.sum(b[u, j] for u in g.neighbor_iterator(v)) - b[v, i], min=0)

    # is_used[i] can be set to 1 only if the color is used
    for i in range(k):
        p.add_constraint(p.sum(b[v, i] for v in g) - is_used[i], min=0)

    # Trying to use as many colors as possible
    p.set_objective(p.sum(is_used[i] for i in range(k)))

    from sage.numerical.mip import MIPSolverException
    try:
        p.solve(log=verbose)
    except MIPSolverException:
        raise ValueError("this graph cannot be colored with k colors")

    from sage.rings.integer import Integer
    is_used = p.get_values(is_used, convert=bool, tolerance=integrality_tolerance)
    obj = Integer(sum(1 for i in range(k) if is_used[i]))

    if value_only:
        return obj

    # Building the dictionary associating its color to every vertex
    b = p.get_values(b, convert=bool, tolerance=integrality_tolerance)
    cdef dict coloring = {}

    for v in g:
        for i in range(k):
            if b[v, i]:
                coloring[v] = i
                break

    return obj, coloring


def b_coloring(g, k, value_only=True, solver=None, verbose=0,
               *, integrality_tolerance=1e-3):
    r"""
    Compute b-chromatic numbers and b-colorings.

    This function computes a b-coloring with at most `k` colors that maximizes
    the number of colors, if such a coloring exists.

    Definition :

    Given a proper coloring of a graph `G` and a color class `C` such that none
    of its vertices have neighbors in all the other color classes, one can
    eliminate color class `C` assigning to each of its elements a missing color
    in its neighborhood.

    Let a b-vertex be a vertex with neighbors in all other colorings. Then, one
    can repeat the above procedure until a coloring is obtained where every
    color class contains a b-vertex, in which case none of the color classes can
    be eliminated with the same idea. So, one can define a b-coloring as a
    proper coloring where each color class has a b-vertex.

    In the worst case, after successive applications of the above procedure, one
    get a proper coloring that uses a number of colors equal to the b-chromatic
    number of `G` (denoted `\chi_b(G)`): the maximum `k` such that `G` admits a
    b-coloring with `k` colors.

    A useful upper bound for calculating the b-chromatic number is the
    following. If `G` admits a b-coloring with `k` colors, then there are `k`
    vertices of degree at least `k - 1` (the b-vertices of each color
    class). So, if we set `m(G) = \max \{k | \text{there are } k \text{ vertices
    of degree at least } k - 1 \}`, we have that `\chi_b(G) \leq m(G)`.


    .. NOTE::

       This method computes a b-coloring that uses at *MOST* `k` colors. If this
       method returns a value equal to `k`, it cannot be assumed that `k` is
       equal to `\chi_b(G)`. Meanwhile, if it returns any value `k' < k`, this
       is a certificate that the Grundy number of the given graph is `k'`.

       As `\chi_b(G)\leq m(G)`, it can be assumed that `\chi_b(G) = k` if
       ``b_coloring(g, k)`` returns `k` when `k = m(G)`.

    INPUT:

    - ``k`` -- integer; maximum number of colors

    - ``value_only`` -- boolean (default: ``True``); when set to ``True``, only
      the number of colors is returned. Otherwise, the pair ``(nb_colors,
      coloring)`` is returned, where ``coloring`` is a dictionary associating
      its color (integer) to each vertex of the graph.

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    ALGORITHM:

    Integer Linear Program.

    EXAMPLES:

    The b-chromatic number of a `P_5` is equal to 3::

        sage: from sage.graphs.graph_coloring import b_coloring
        sage: g = graphs.PathGraph(5)
        sage: b_coloring(g, 5)                                                          # needs sage.numerical.mip
        3

    The b-chromatic number of the Petersen Graph is equal to 3::

        sage: g = graphs.PetersenGraph()
        sage: b_coloring(g, 5)                                                          # needs sage.numerical.mip
        3

    It would have been sufficient to set the value of ``k`` to 4 in this case,
    as `4 = m(G)`.
    """
    g._scream_if_not_simple(allow_multiple_edges=True)

    # Calculate the upper bound m(G)
    # To do so, it takes the list of degrees in non-increasing order and
    # computes the largest i, such that the ith degree on the list is at least
    # i - 1 (note that in the code we need to take in consideration that the
    # indices of the list starts with 0)

    cdef list deg = g.degree_sequence()
    cdef int n = g.order()
    for i in range(n):
        if deg[i] < i:
            break
    if i != n - 1:
        m = i
    else:
        m = n

    # In case the k specified by the user is greater than m(G), make k = m(G)
    if k > m:
        k = m

    p = MixedIntegerLinearProgram(solver=solver)

    # color[v,i] is set to 1 if and only if v is colored i
    color = p.new_variable(binary=True)

    # b[v,i] is set to 1 if and only if v is a b-vertex from color class i
    b = p.new_variable(binary=True)

    # is_used[i] is set to 1 if and only if color[i] is used by some vertex
    is_used = p.new_variable(binary=True)

    # Each vertex is in exactly one class
    for v in g:
        p.add_constraint(p.sum(color[v, i] for i in range(k)), min=1, max=1)

    # Adjacent vertices have distinct colors
    for u, v in g.edge_iterator(labels=None):
        for i in range(k):
            p.add_constraint(color[u, i] + color[v, i], max=1)

    # The following constraints ensure that if v is a b-vertex of color i
    # then it has a neighbor colored j for every j != i

    for v in g:
        for i in range(k):
            for j in range(k):
                if j != i:
                    # If v is not a b-vertex of color i, the constraint
                    # is always satisfied, since the only possible
                    # negative term in this case is -is_used[j] which is
                    # cancelled by + 1. If v is a b-vertex of color i
                    # then we MUST have sum(color[w,j] for w in g.neighbors(v))
                    # valued at least 1, which means that v has a neighbour in
                    # color j, as desired.
                    p.add_constraint(p.sum(color[w, j] for w in g.neighbor_iterator(v))
                                     - b[v, i] + 1 - is_used[j], min=0)

    # if color i is used, there is a vertex colored i
    for i in range(k):
        p.add_constraint(p.sum(color[v, i] for v in g) - is_used[i], min=0)

    # if there is a vertex colored with color i, then i is used
    for v in g:
        for i in range(k):
            p.add_constraint(color[v, i] - is_used[i], max=0)

    # a color class is used if and only if it has one b-vertex
    for i in range(k):
        p.add_constraint(p.sum(b[w, i] for w in g) - is_used[i], min=0, max=0)

    # We want to maximize the number of used colors
    p.set_objective(p.sum(is_used[i] for i in range(k)))

    from sage.numerical.mip import MIPSolverException
    try:
        p.solve(log=verbose)
    except MIPSolverException:
        raise ValueError("this graph cannot be colored with k colors")

    from sage.rings.integer import Integer
    is_used = p.get_values(is_used, convert=bool, tolerance=integrality_tolerance)
    obj = Integer(sum(1 for i in range(k) if is_used[i]))

    if value_only:
        return obj

    # Building the dictionary associating its color to every vertex
    c = p.get_values(color, convert=bool, tolerance=integrality_tolerance)
    cdef dict coloring = {}

    for v in g:
        for i in range(k):
            if c[v, i]:
                coloring[v] = i
                break

    return obj, coloring


def edge_coloring(g, value_only=False, vizing=False, hex_colors=False, solver=None, verbose=0,
                  *, integrality_tolerance=1e-3):
    r"""
    Compute chromatic index and edge colorings.

    INPUT:

    - ``g`` -- a graph

    - ``value_only`` -- boolean (default: ``False``):

      - When set to ``True``, only the chromatic index is returned

      - When set to ``False``, a partition of the edge set into matchings is
        returned if possible

    - ``vizing`` -- boolean (default: ``False``):

      - When set to ``True``, finds an edge coloring using the algorithm
        described at [MG1992]_. This always results in a coloring with `\Delta + 1`
        colors, where `\Delta` is equal to the maximum degree in the graph, even
        if one of the colors is empty, for the sake of consistency.

      - When set to ``False``, tries to find a `\Delta`-edge-coloring using
        Mixed Integer Linear Programming (MILP). If impossible, returns a
        `(\Delta + 1)`-edge-coloring.  Please note that determining if the
        chromatic index of a graph equals `\Delta` is computationally difficult,
        and could take a long time.

    - ``hex_colors`` -- boolean (default: ``False``); when set to ``True``, the
      partition returned is a dictionary whose keys are colors and whose values
      are the color classes (ideal for plotting)

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    OUTPUT:

    In the following, `\Delta` is equal to the maximum degree in the graph
    ``g``.

    - If ``vizing=True`` and ``value_only=False``, return a partition of the
      edge set into `\Delta + 1` matchings.

    - If ``vizing=False`` and ``value_only=True``, return the chromatic index.

    - If ``vizing=False`` and ``value_only=False``, return a partition of the
      edge set into the minimum number of matchings.

    - If ``vizing=True`` and ``value_only=True``, should return something, but
      mainly you are just trying to compute the maximum degree of the graph, and
      this is not the easiest way. By Vizing's theorem, a graph has a chromatic
      index equal to `\Delta` or to `\Delta + 1`.

    .. NOTE::

       In a few cases, it is possible to find very quickly the chromatic index
       of a graph, while it remains a tedious job to compute a corresponding
       coloring. For this reason, ``value_only = True`` can sometimes be much
       faster, and it is a bad idea to compute the whole coloring if you do not
       need it !

    .. SEEALSO::

        - :wikipedia:`Edge_coloring` for further details on edge coloring
        - :meth:`~Graph.chromatic_index`
        - :meth:`~Graph.fractional_chromatic_index`
        - :meth:`~Graph.chromatic_number`
        - :meth:`sage.graphs.graph_coloring.vertex_coloring`

    EXAMPLES:

    The Petersen graph has chromatic index 4::

        sage: # needs sage.numerical.mip
        sage: from sage.graphs.graph_coloring import edge_coloring
        sage: g = graphs.PetersenGraph()
        sage: edge_coloring(g, value_only=True, solver='GLPK')
        4
        sage: color_classes = edge_coloring(g, value_only=False, solver='GLPK')
        sage: len(color_classes)
        4
        sage: len(set(frozenset(e) for C in color_classes for e in C)) == g.size()
        True
        sage: all(g.has_edge(e) for C in color_classes for e in C)
        True
        sage: all(len(Graph(C).matching()) == len(C) for C in color_classes)            # needs networkx
        True
        sage: color_classes = edge_coloring(g, value_only=False,                        # needs sage.plot
        ....:                               hex_colors=True, solver='GLPK')
        sage: sorted(color_classes.keys())                                              # needs sage.plot
        ['#00ffff', '#7f00ff', '#7fff00', '#ff0000']

    Complete graphs are colored using the linear-time round-robin coloring::

        sage: from sage.graphs.graph_coloring import edge_coloring
        sage: len(edge_coloring(graphs.CompleteGraph(20)))                              # needs sage.numerical.mip
        19

    The chromatic index of a non connected graph is the maximum over its
    connected components::

        sage: g = graphs.CompleteGraph(4) + graphs.CompleteGraph(10)
        sage: edge_coloring(g, value_only=True)                                         # needs sage.numerical.mip
        9

    TESTS:

    Graph without edge::

        sage: g = Graph(2)
        sage: edge_coloring(g)                                                          # needs sage.numerical.mip
        []
        sage: edge_coloring(g, value_only=True)                                         # needs sage.numerical.mip
        0
        sage: edge_coloring(g, hex_colors=True)                                         # needs sage.numerical.mip
        {}
    """
    g._scream_if_not_simple()

    if not g.order() or not g.size():
        if value_only:
            return 0
        return dict() if hex_colors else list()

    # The chromatic index of g is the maximum value over its connected
    # components, and the edge coloring is the union of the edge
    # coloring of its connected components
    cdef list L = [g] if g.is_connected() else g.connected_components_subgraphs()
    cdef int chi = 0
    cdef list classes = [], vertices

    if vizing:
        classes = _vizing_edge_coloring(g)
        if len(classes) == max(g.degree()):
            # guaranteeing that vizing=True always returns Delta+1 colors
            # for backward compatibility
            classes.append([])
    else:
        def extend_color_classes(classes, coloring):
            # create missing color classes, if any
            for _ in range(len(classes), len(coloring)):
                classes.append([])
            # add edges to classes
            for i, edges in enumerate(coloring):
                classes[i].extend(edges)

        for h in L:

            if not h.size():
                continue

            # We get the vertex of maximum degree and its degree
            Delta, X = max([(d, v) for v, d in h.degree_iterator(labels=True)], key=lambda x: x[0])

            if Delta + 1 <= chi:
                c = _vizing_edge_coloring(h)
                extend_color_classes(classes, c)
                continue

            if value_only:
                if Delta + 1 <= chi:
                    continue
                if h.is_overfull():
                    chi = max(chi, Delta + 1)
                    continue

            if h.is_clique():
                if value_only:
                    chi = max(chi, h.order() if h.order() % 2 else (h.order() - 1))
                    continue
                vertices = list(h)
                r = round_robin(h.order())
                # create missing color classes, if any
                for i in range(len(classes), max(r.edge_labels()) + 1):
                    classes.append([])
                # add edges to classes
                r.relabel(perm=vertices, inplace=True)
                for u, v, c in r.edge_iterator():
                    classes[c].append((u, v))
                continue

            p = MixedIntegerLinearProgram(maximization=True, solver=solver)
            color = p.new_variable(binary=True)
            # A vertex cannot have two incident edges with the same color.
            for v in h:
                for i in range(Delta):
                    p.add_constraint(p.sum(color[frozenset((u, v)), i] for u in h.neighbor_iterator(v)) <= 1)
            # An edge must have a color
            for u, v in h.edge_iterator(labels=False):
                p.add_constraint(p.sum(color[frozenset((u, v)), i] for i in range(Delta)) == 1)
            # We color the edges of the vertex of maximum degree
            for i, v in enumerate(h.neighbor_iterator(X)):
                p.add_constraint(color[frozenset((v, X)), i] == 1)

            from sage.numerical.mip import MIPSolverException
            try:
                p.solve(objective_only=value_only, log=verbose)
            except MIPSolverException:
                # The coloring fails with Delta colors
                chi = max(chi, Delta + 1)
                if not value_only:
                    c = _vizing_edge_coloring(h)
                    extend_color_classes(classes, c)
                continue

            chi = max(chi, Delta)
            if not value_only:
                # create missing color classes, if any
                for i in range(len(classes), Delta):
                    classes.append([])
                # add edges to color classes
                color = p.get_values(color, convert=bool, tolerance=integrality_tolerance)
                for e in h.edge_iterator(labels=False):
                    fe = frozenset(e)
                    for i in range(Delta):
                        if color[fe, i]:
                            classes[i].append(e)
                            break

        if value_only:
            return chi

    # if needed, builds a dictionary from the color classes adding colors
    return format_coloring(classes, value_only=not hex_colors, hex_colors=hex_colors)


def _vizing_edge_coloring(g):
    r"""
    Compute an edge coloring with at most `\Delta + 1` colors.

    INPUT:

    - ``g`` -- a graph

    OUTPUT: a partition of the edge set into at most `\Delta + 1` matchings

    .. SEEALSO::

        - :wikipedia:`Edge_coloring` for further details on edge coloring
        - :wikipedia:`Vizing's_theorem` for further details on Vizing's theorem

    ALGORITHM:

    This function's implementation is based on the algorithm described at [MG1992]_

    EXAMPLES:

    Coloring the edges of the Petersen Graph::

       sage: from sage.graphs.graph_coloring import _vizing_edge_coloring
       sage: g = graphs.PetersenGraph()
       sage: color_classes = _vizing_edge_coloring(g)
       sage: len(color_classes) == max(g.degree()) + 1
       True
       sage: len(set(frozenset(e) for C in color_classes for e in C)) == g.size()
       True
       sage: all(g.has_edge(e) for C in color_classes for e in C)
       True
       sage: all(len(Graph(C).matching()) == len(C) for C in color_classes)             # needs networkx
       True

    Coloring the edges of the Star Graph::

       sage: from sage.graphs.graph_coloring import _vizing_edge_coloring
       sage: g = graphs.StarGraph(5)
       sage: len(_vizing_edge_coloring(g))
       5

    TESTS:

    Graph without edge::

       sage: g = Graph(2)
       sage: _vizing_edge_coloring(g)
       []

    Random graphs::

       sage: from sage.graphs.generators.random import RandomGNP
       sage: g = RandomGNP(randint(1, 20), random())
       sage: colors = _vizing_edge_coloring(g)
       sage: Delta = max(g.degree())
       sage: len(colors) in [Delta, Delta + 1]
       True
       sage: len(set(frozenset(e) for C in colors for e in C)) == g.size()
       True
       sage: all(g.has_edge(e) for C in colors for e in C)
       True
       sage: all(len(Graph(C).matching()) == len(C) for C in colors)                    # needs networkx
       True
    """
    # This implementation was discussed in github issue #34809

    # dictionary mapping edges to colors
    e_colors = {frozenset(e): None for e in g.edge_iterator(labels=False, sort_vertices=False)}

    # finds every color adjacent to vertex v
    def colors_of(v):
        colors = {e_colors[frozenset((u, v))] for u in g.neighbor_iterator(v)}
        colors.discard(None)
        return colors

    # constructs a maximal fan <f..l> of X where X is edge[0] and f is edge[1]
    def maximal_fan(edge):
        fan_center, rear = edge
        cdef set rear_colors = colors_of(rear)
        cdef list neighbors = [n for n in g.neighbor_iterator(fan_center)
                                   if e_colors[frozenset((n, fan_center))] is not None]
        cdef list fan = [rear]
        cdef bint can_extend_fan = True
        while can_extend_fan:
            can_extend_fan = False
            new_neighbors = []
            for n in neighbors:
                if e_colors[frozenset((fan_center, n))] not in rear_colors:
                    fan.append(n)
                    rear = n
                    rear_colors = colors_of(rear)
                    can_extend_fan = True
                else:
                    new_neighbors.append(n)
            neighbors = new_neighbors
        return fan

    # gives each edge Xu in the fan <f..w> the color of Xu+ and uncolors Xw
    def rotate_fan(fan_center, fan):
        for i in range(1, len(fan)):
            e_colors[frozenset((fan_center, fan[i - 1]))] = e_colors[frozenset((fan_center, fan[i]))]
        e_colors[frozenset((fan_center, fan[-1]))] = None

    # computes the maximal ab-path starting at v
    def find_path(v, a, b):
        cdef list path = [v]
        cdef bint stop = False
        while not stop:
            stop = True
            for u in g.neighbor_iterator(v):
                if e_colors[frozenset((u, v))] == a:
                    path.append(u)
                    # exchange the role of colors a and b and go to next vertex
                    a, b = b, a
                    v = u
                    stop = False
                    break
        return path

    # exchanges the color of every edge on the ab-path starting at v
    def invert_path(v, a, b):
        cdef list path = find_path(v, a, b)
        for e in zip(path[:-1], path[1:]):
            f = frozenset(e)
            e_colors[f] = a if e_colors[f] == b else b

    # returns the ´smallest´ color free at v
    def find_free_color(v):
        colors = colors_of(v)
        for c in range(g.degree(v) + 1):
            if c not in colors:
                return c

    for e in g.edge_iterator(labels=False, sort_vertices=False):
        fan = maximal_fan(e)
        fan_center = e[0]
        rear = fan[-1]
        c = find_free_color(fan_center)
        d = find_free_color(rear)
        invert_path(fan_center, d, c)
        for i, v in enumerate(fan):
            if d not in colors_of(v):
                fan = fan[:i + 1]
                break
        rotate_fan(fan_center, fan)
        e_colors[frozenset((fan_center, fan[-1]))] = d

    matchings = {}
    for edge, c in e_colors.items():
        matchings[c] = matchings.get(c, []) + [tuple(edge)]
    classes = list(matchings.values())

    return classes


def round_robin(n):
    r"""
    Compute a round-robin coloring of the complete graph on `n` vertices.

    A round-robin coloring of the complete graph `G` on `2n` vertices
    (`V = [0, \dots, 2n - 1]`) is a proper coloring of its edges such that
    the edges with color `i` are all the `(i + j, i - j)` plus the
    edge `(2n - 1, i)`.

    If `n` is odd, one obtain a round-robin coloring of the complete graph
    through the round-robin coloring of the graph with `n + 1` vertices.

    INPUT:

    - ``n`` -- the number of vertices in the complete graph

    OUTPUT:

    - A :meth:`~sage.graphs.graph_generators.GraphGenerators.CompleteGraph` with
      labelled edges such that the label of each edge is its color.

    EXAMPLES::

        sage: from sage.graphs.graph_coloring import round_robin
        sage: round_robin(3).edges(sort=True)
        [(0, 1, 2), (0, 2, 1), (1, 2, 0)]

    ::

        sage: round_robin(4).edges(sort=True)
        [(0, 1, 2), (0, 2, 1), (0, 3, 0), (1, 2, 0), (1, 3, 1), (2, 3, 2)]


    For higher orders, the coloring is still proper and uses the expected
    number of colors::

        sage: g = round_robin(9)
        sage: sum(Set(e[2] for e in g.edges_incident(v)).cardinality() for v in g) == 2 * g.size()
        True
        sage: Set(e[2] for e in g.edge_iterator()).cardinality()
        9

    ::

        sage: g = round_robin(10)
        sage: sum(Set(e[2] for e in g.edges_incident(v)).cardinality() for v in g) == 2 * g.size()
        True
        sage: Set(e[2] for e in g.edge_iterator()).cardinality()
        9
    """
    if n <= 1:
        raise ValueError("there must be at least two vertices in the graph")

    def my_mod(x, y):
        return x - y * (x // y)
    if not n % 2:
        from sage.graphs.generators.basic import CompleteGraph
        g = CompleteGraph(n)
        for i in range(n - 1):
            g.set_edge_label(n - 1, i, i)
            for j in range(1, (n - 1) // 2 + 1):
                g.set_edge_label(my_mod(i - j, n - 1), my_mod(i + j, n - 1), i)
    else:
        g = round_robin(n + 1)
        g.delete_vertex(n)
    return g


def linear_arboricity(g, plus_one=None, hex_colors=False, value_only=False,
                      solver=None, verbose=0, *, integrality_tolerance=1e-3):
    r"""
    Compute the linear arboricity of the given graph.

    The linear arboricity of a graph `G` is the least number `la(G)` such that
    the edges of `G` can be partitioned into linear forests (i.e. into forests
    of paths).

    Obviously, `la(G)\geq \left\lceil \frac{\Delta(G)}{2} \right\rceil`.

    It is conjectured in [Aki1980]_ that `la(G)\leq \left\lceil
    \frac{\Delta(G)+1}{2} \right\rceil`.

    INPUT:

    - ``plus_one`` -- integer (default: ``None``); whether to use `\left\lceil
      \frac{\Delta(G)}{2} \right\rceil` or `\left\lceil \frac{\Delta(G)+1}{2}
      \right\rceil` colors.

      - If ``0``, computes a decomposition of `G` into `\left\lceil
        \frac{\Delta(G)}{2} \right\rceil` forests of paths

      - If ``1``, computes a decomposition of `G` into `\left\lceil
        \frac{\Delta(G)+1}{2} \right\rceil` colors, which is the conjectured
        general bound.

      - If ``plus_one = None`` (default), computes a decomposition using the
        least possible number of colors.

    - ``hex_colors`` -- boolean (default: ``False``):

      - If ``hex_colors = True``, the function returns a dictionary associating
        to each color a list of edges (meant as an argument to the
        ``edge_colors`` keyword of the ``plot`` method).

      - If ``hex_colors = False`` (default value), returns a list of graphs
        corresponding to each color class.

    - ``value_only`` -- boolean (default: ``False``):

      - If ``value_only = True``, only returns the linear arboricity as an
        integer value.

      - If ``value_only = False``, returns the color classes according to the
        value of ``hex_colors``

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    ALGORITHM:

    Linear Programming

    COMPLEXITY:

    NP-Hard

    EXAMPLES:

    Obviously, a square grid has a linear arboricity of 2, as the set of
    horizontal lines and the set of vertical lines are an admissible partition::

        sage: from sage.graphs.graph_coloring import linear_arboricity
        sage: g = graphs.Grid2dGraph(4, 4)                                              # needs sage.numerical.mip
        sage: g1,g2 = linear_arboricity(g)                                              # needs sage.numerical.mip

    Each graph is of course a forest::

        sage: g1.is_forest() and g2.is_forest()                                         # needs sage.numerical.mip
        True

    Of maximum degree 2::

        sage: max(g1.degree()) <= 2 and max(g2.degree()) <= 2                           # needs sage.numerical.mip
        True

    Which constitutes a partition of the whole edge set::

        sage: all((g1.has_edge(e) or g2.has_edge(e))                                    # needs sage.numerical.mip
        ....:     for e in g.edge_iterator(labels=None))
        True

    TESTS:

    Asking for the value of the linear arboricity only (:issue:`24991`)::

        sage: from sage.graphs.graph_coloring import linear_arboricity
        sage: sorted(linear_arboricity(G, value_only=True) for G in graphs(4))          # needs sage.numerical.mip
        [0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2]

    Test parameter ``hex_color`` (:issue:`26228`)::

        sage: from sage.graphs.graph_coloring import linear_arboricity
        sage: g = graphs.Grid2dGraph(4, 4)
        sage: d = linear_arboricity(g, hex_colors=True)                                 # needs sage.numerical.mip sage.plot
        sage: sorted(d)                                                                 # needs sage.numerical.mip sage.plot
        ['#00ffff', '#ff0000']
    """
    g._scream_if_not_simple()
    from sage.rings.integer import Integer

    if plus_one is None:
        try:
            return linear_arboricity(g,
                                     plus_one=0,
                                     value_only=value_only,
                                     hex_colors=hex_colors,
                                     solver=solver,
                                     verbose=verbose,
                                     integrality_tolerance=integrality_tolerance)
        except ValueError:
            return linear_arboricity(g,
                                     plus_one=1,
                                     value_only=value_only,
                                     hex_colors=hex_colors,
                                     solver=solver,
                                     verbose=verbose,
                                     integrality_tolerance=integrality_tolerance)
    elif plus_one == 1:
        k = (Integer(1 + max(g.degree())) / 2).ceil()
    elif not plus_one:
        k = (Integer(max(g.degree())) / 2).ceil()
    else:
        raise ValueError("plus_one must be equal to 0,1, or to None!")

    p = MixedIntegerLinearProgram(solver=solver)

    # c is a boolean value such that c[i,(u,v)] = 1 if and only if (u,v) is
    # colored with i
    c = p.new_variable(binary=True)

    # relaxed value
    r = p.new_variable(nonnegative=True)

    MAD = 1 - 1 / (Integer(g.order()) * 2)

    # Partition of the edges
    for u, v in g.edge_iterator(labels=None):
        p.add_constraint(p.sum(c[i, frozenset((u, v))] for i in range(k)), max=1, min=1)

    for i in range(k):

        # r greater than c
        for u, v in g.edge_iterator(labels=None):
            p.add_constraint(r[i, (u, v)] + r[i, (v, u)] - c[i, frozenset((u, v))], max=0, min=0)

        # Maximum degree 2
        for u in g:
            p.add_constraint(p.sum(c[i, frozenset((u, v))] for v in g.neighbor_iterator(u)), max=2)

            # no cycles
            p.add_constraint(p.sum(r[i, (u, v)] for v in g.neighbor_iterator(u)), max=MAD)

    from sage.numerical.mip import MIPSolverException
    try:
        p.solve(objective_only=value_only, log=verbose)
        if value_only:
            return k

    except MIPSolverException:
        if plus_one:
            raise RuntimeError("It looks like you have found a counterexample "
                               "to a very old conjecture. Please do not loose "
                               "it ! Please publish it, and send a post to "
                               "sage-devel to warn us. We implore you!")
        else:
            raise ValueError("this graph cannot be colored with the given number of colors")

    c = p.get_values(c, convert=bool, tolerance=integrality_tolerance)

    cdef list answer

    if hex_colors:
        answer = [[] for i in range(k)]

        def add(uv, i):
            return answer[i].append(uv)
    else:
        gg = copy(g)
        gg.delete_edges(g.edge_iterator())
        answer = [copy(gg) for i in range(k)]

        def add(uv, i):
            return answer[i].add_edge(uv)

    for i in range(k):
        for u, v in g.edge_iterator(labels=None):
            if c[i, frozenset((u, v))]:
                add((u, v), i)

    return format_coloring(answer, value_only=not hex_colors, hex_colors=hex_colors)


def acyclic_edge_coloring(g, hex_colors=False, value_only=False, k=0,
                          solver=None, verbose=0, *, integrality_tolerance=1e-3):
    r"""
    Compute an acyclic edge coloring of the current graph.

    An edge coloring of a graph is a assignment of colors to the edges of a
    graph such that :

    - the coloring is proper (no adjacent edges share a color)
    - For any two colors `i,j`, the union of the edges colored with `i` or `j`
      is a forest.

    The least number of colors such that such a coloring exists for a graph `G`
    is written `\chi'_a(G)`, also called the acyclic chromatic index of `G`.

    It is conjectured that this parameter cannot be too different from the
    obvious lower bound `\Delta(G) \leq \chi'_a(G)`, `\Delta(G)` being the
    maximum degree of `G`, which is given by the first of the two constraints.
    Indeed, it is conjectured that `\Delta(G)\leq \chi'_a(G)\leq \Delta(G) + 2`.

    INPUT:

    - ``hex_colors`` -- boolean (default: ``False``):

      - If ``hex_colors = True``, the function returns a dictionary associating
        to each color a list of edges (meant as an argument to the
        ``edge_colors`` keyword of the ``plot`` method).

      - If ``hex_colors = False`` (default value), returns a list of graphs
        corresponding to each color class.

    - ``value_only`` -- boolean (default: ``False``):

      - If ``value_only = True``, only returns the acyclic chromatic index as an
        integer value

      - If ``value_only = False``, returns the color classes according to the
        value of ``hex_colors``

    - ``k`` -- integer; the number of colors to use

      - If ``k > 0``, computes an acyclic edge coloring using `k` colors.

      - If ``k = 0`` (default), computes a coloring of `G` into `\Delta(G) + 2`
        colors, which is the conjectured general bound.

      - If ``k = None``, computes a decomposition using the least possible
        number of colors.

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    ALGORITHM:

    Linear Programming

    EXAMPLES:

    The complete graph on 8 vertices cannot be acyclically edge-colored with
    less `\Delta + 1` colors, but it can be colored with `\Delta + 2 = 9`::

        sage: from sage.graphs.graph_coloring import acyclic_edge_coloring
        sage: g = graphs.CompleteGraph(8)
        sage: colors = acyclic_edge_coloring(g)                                         # needs sage.numerical.mip

    Each color class is of course a matching ::

        sage: all(max(gg.degree()) <= 1 for gg in colors)                               # needs sage.numerical.mip
        True

    These matchings being a partition of the edge set::

        sage: all(any(gg.has_edge(e) for gg in colors)                                  # needs sage.numerical.mip
        ....:     for e in g.edge_iterator(labels=False))
        True

    Besides, the union of any two of them is a forest ::

        sage: all(g1.union(g2).is_forest() for g1 in colors for g2 in colors)           # needs sage.numerical.mip
        True

    If one wants to acyclically color a cycle on `4` vertices, at least 3 colors
    will be necessary. The function raises an exception when asked to color it
    with only 2::

        sage: g = graphs.CycleGraph(4)
        sage: acyclic_edge_coloring(g, k=2)                                             # needs sage.numerical.mip
        Traceback (most recent call last):
        ...
        ValueError: this graph cannot be colored with the given number of colors

    The optimal coloring give us `3` classes::

        sage: colors = acyclic_edge_coloring(g, k=None)                                 # needs sage.numerical.mip
        sage: len(colors)                                                               # needs sage.numerical.mip
        3

    TESTS:

    Issue :issue:`24991` is fixed::

        sage: from sage.graphs.graph_coloring import acyclic_edge_coloring
        sage: sorted(acyclic_edge_coloring(G, value_only=True) for G in graphs(4))      # needs sage.numerical.mip
        [2, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5]

    Test parameter ``hex_color`` (:issue:`26228`)::

        sage: from sage.graphs.graph_coloring import acyclic_edge_coloring
        sage: g = graphs.CompleteGraph(4)
        sage: d = acyclic_edge_coloring(g, hex_colors=True)                             # needs sage.numerical.mip sage.plot
        sage: sorted(d)                                                                 # needs sage.numerical.mip sage.plot
        ['#0066ff', '#00ff66', '#cbff00', '#cc00ff', '#ff0000']

    The acyclic chromatic index of a graph without edge is 0 (:issue:`27079`)::

        sage: from sage.graphs.graph_coloring import acyclic_edge_coloring
        sage: g = Graph(3)
        sage: acyclic_edge_coloring(g, k=None, value_only=True)                         # needs sage.numerical.mip
        0
        sage: acyclic_edge_coloring(g, k=None, hex_colors=True)                         # needs sage.numerical.mip sage.plot
        {}
        sage: acyclic_edge_coloring(g, k=None, hex_colors=False)                        # needs sage.numerical.mip sage.plot
        []

    Empty graph  (:issue:`27079`)::

        sage: from sage.graphs.graph_coloring import acyclic_edge_coloring
        sage: acyclic_edge_coloring(Graph(), k=None, value_only=True)                   # needs sage.numerical.mip
        0
    """
    g._scream_if_not_simple(allow_multiple_edges=True)

    from sage.rings.integer import Integer
    from sage.combinat.subset import Subsets

    if not g.order() or not g.size():
        if k == 0:
            k = 2
        if value_only:
            return 0 if k is None else k
        else:
            if k is None:
                return {} if hex_colors else []
            if hex_colors:
                return format_coloring([[] for _ in range(k)], hex_colors=True)
            else:
                return [copy(g) for _ in range(k)]

    if k is None:
        k = max(g.degree())

        while True:
            try:
                return acyclic_edge_coloring(g,
                                             value_only=value_only,
                                             hex_colors=hex_colors,
                                             k=k,
                                             solver=solver,
                                             verbose=verbose,
                                             integrality_tolerance=integrality_tolerance)
            except ValueError:
                k += 1

        raise RuntimeError("this should not happen, please report a bug!")

    elif not k:
        k = max(g.degree()) + 2

    p = MixedIntegerLinearProgram(solver=solver)

    # c is a binary variable such that c[i,(u,v)] = 1 if and only if (u,v) is
    # colored with i
    c = p.new_variable(binary=True)

    # relaxed value
    r = p.new_variable(nonnegative=True)

    def E(x, y):
        return frozenset((x, y))

    MAD = 1 - 1/(Integer(g.order()) * 2)

    # Partition of the edges: each edge is assigned a unique color
    for u, v in g.edge_iterator(labels=None):
        p.add_constraint(p.sum(c[i, E(u, v)] for i in range(k)), max=1, min=1)

    for i in range(k):

        # Maximum degree 1
        for u in g:
            p.add_constraint(p.sum(c[i, E(u, v)] for v in g.neighbor_iterator(u)), max=1)

    for i, j in Subsets(range(k), 2):
        # r is greater than c
        for u in g:
            p.add_constraint(p.sum(r[(i, j), (u, v)] for v in g.neighbor_iterator(u)), max=MAD)

        # r greater than c
        for u, v in g.edge_iterator(labels=None):
            p.add_constraint(r[(i, j), (u, v)] + r[(i, j), (v, u)] - c[i, E(u, v)] - c[j, E(u, v)], max=0, min=0)

    p.set_objective(None)

    from sage.numerical.mip import MIPSolverException
    try:
        p.solve(objective_only=value_only, log=verbose)
        if value_only:
            return k

    except MIPSolverException:
        if k == max(g.degree()) + 2:
            raise RuntimeError("It looks like you have found a counterexample to "
                               "a very old conjecture. Please do not loose it ! "
                               "Please publish it, and send a post to sage-devel "
                               "to warn us. We implore you!")
        else:
            raise ValueError("this graph cannot be colored with the given number of colors")

    c = p.get_values(c, convert=bool, tolerance=integrality_tolerance)

    if hex_colors:
        answer = [[] for i in range(k)]

        def add(uv, i):
            return answer[i].append(uv)
    else:
        gg = copy(g)
        gg.delete_edges(g.edge_iterator())
        answer = [copy(gg) for i in range(k)]

        def add(uv, i):
            return answer[i].add_edge(uv)

    for i in range(k):
        for u, v in g.edge_iterator(labels=None):
            if c[i, E(u, v)]:
                add((u, v), i)

    return format_coloring(answer, value_only=not hex_colors, hex_colors=hex_colors)


cdef class Test:
    r"""
    This class performs randomized testing for :func:`all_graph_colorings`.

    Since everything else in this file is derived from :func:`all_graph_colorings`, this
    is a pretty good randomized tester for the entire file.  Note that for a
    graph `G`, ``G.chromatic_polynomial()`` uses an entirely different
    algorithm, so we provide a good, independent test.
    """

    def random(self, tests=1000):
        r"""
        Call ``self.random_all_graph_colorings()``.

        In the future, if other methods are added, it should call them, too.

        TESTS::

            sage: from sage.graphs.graph_coloring import Test
            sage: Test().random(1)                                                      # needs cliquer sage.libs.flint
        """
        self.random_all_graph_colorings(tests)

    def random_all_graph_colorings(self, tests=2):
        r"""
        Verify the results of ``all_graph_colorings()`` in three ways:

        #. all colorings are unique

        #. number of m-colorings is `P(m)` (where `P` is the chromatic
           polynomial of the graph being tested)

        #. colorings are valid -- that is, that no two vertices of the same
           color share an edge.

        TESTS::

            sage: from sage.graphs.graph_coloring import Test
            sage: Test().random_all_graph_colorings(1)                                  # needs cliquer sage.libs.flint
        """
        from sage.graphs.generators.random import RandomGNP
        cdef set S
        cdef list parts
        for _ in range(tests):
            G = RandomGNP(10, .5)
            Q = G.chromatic_polynomial()
            chi = G.chromatic_number()

            S = set()

            for C in all_graph_colorings(G, chi):
                parts = [C[k] for k in C]
                for P in parts:
                    lenP = len(P)
                    for i in range(lenP):
                        for j in range(i + 1, lenP):
                            if G.has_edge(P[i], P[j]):
                                raise RuntimeError("coloring failed")

                # make the dict into a frozenset for quick uniqueness checking
                S.add(frozenset((k, frozenset(C[k])) for k in C))

            if len(S) != Q(chi):
                raise RuntimeError("incorrect number of unique colorings")



# sage_setup: distribution = sagemath-graphs
# cython: binding=True
# distutils: language = c++
# distutils: extra_compile_args = -std=c++11
r"""
Slice decomposition

This module implements an extended lexBFS algorithm for computing the slice
decomposition of undirected graphs and the class :class:`~SliceDecomposition` to
represent such decompositions.

A formal definition of slice decompositions can be found in Section 3.2 of
[TCHP2008]_ and a description of the extended lexBFS algorithm is given in
Section 3.3 of [TCHP2008]_.

AUTHORS:

- Cyril Bouvier (2024-06-25): initial version
"""
# ****************************************************************************
# Copyright (C) 2024 Cyril Bouvier <cyril.bouvier@lirmm.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from libcpp.algorithm cimport swap
from cython.operator cimport dereference as deref

from sage.graphs.base.c_graph cimport CGraphBackend
from sage.data_structures.bitset_base cimport bitset_in

cdef void extended_lex_BFS(
        CGraph cg, vector[int] &sigma, vector[int] *sigma_inv,
        int initial_v_int, vector[int] *pred, vector[size_t] *xslice_len,
        vector[vector[int]] *lex_label) except *:
    r"""
    Perform a extended lexicographic breadth first search (LexBFS) on the
    undirected graph `G`.

    In addition to computing a LexBFS ordering, the extended LexBFS algorithm
    can be used to compute the slice decomposition of the graph.

    This function implements the `O(n+m)` time algorithm proposed in [HMPV2000]_
    and [TCHP2008]_.

    INPUT:

    - ``cg`` -- a ``CGraph``. This function ignores loops and multiple edges and
      assumes that the graph is undirected.

    - ``sigma`` -- vector of ``int`` to store the ordering of the vertices
      resulting from the LexBFS traversal. At the end, the vector will have size
      `n` (the number of vertices of the graph).

    - ``sigma_inv`` -- a pointer to a vector to store the inverse of the
      permutation ``sigma``. ``sigma_inv`` can be ``NULL`` if the caller does
      not need it (but, note that, the inverse of ``sigma`` is still needed by
      the algorithm, so it does not save time nor memory to have ``sigma_inv``
      equal to ``NULL``). At the end, if ``sigma_inv`` is not NULL, the vector
      pointer by it will have size `n` (the number of vertices of the graph)
      and will satisfy:
        * sigma[deref(sigma_inv)[v_int]] = v_int
        * deref(sigma_inv)[sigma[i]] = i

    - ``initial_v_int`` -- the first vertex to consider. It can be `-1`; in this
      case the first active vertex (corresponding to the first bit set in
      ``cg.active_vertices``) will be taken as first vertex.

    - ``pred`` -- a pointer to a vector of int to store the predecessor of a
      vertex in the LexBFS traversal. ``pred`` can be ``NULL`` if the caller
      does not need it (and the information will not be computed by the
      algorithm). At the end, if ``pred`` is not NULL, the vector pointer by it
      will have size `n` (the number of vertices of the graph) and pred[i] will
      be either -1 (if sigma[i] as no predecessor) or a positive value less than
      n such that the predecessor of sigma[i] is sigma[pred[i]].

    - ``xslice_len`` -- a pointer to a vector of size_t to store the length of
      the x-slices associated with the lexBFS traversal. ``xslice_len`` can be
      ``NULL`` if the caller does not need it (and the information will not be
      computed by the algorithm). At the end, if ``xslice_len`` is not NULL, the
      vector pointer by it will have size `n` (the number of vertices of the
      graph) and the length of the x-slice starting at sigma[i] will be
      xslice_len[i].

    - ``lex_label`` -- a pointer to a vector of vector[int] to store the
      lexicographic labels associated with the lexBFS traversal. ``lex_label``
      can be ``NULL`` if the caller does not need it (and the information will
      not be computed by the algorithm). At the end, if ``lex_label`` is not
      NULL, the vector pointer by it will have size `n` (the number of
      vertices of the graph) and the lexicographic label of sigma[i]
      will given by lex_label[i].

    ALGORITHM:

    This algorithm uses the notion of *partition refinement* to determine the
    exact position of the vertices in the ordering.

    Consider an ordering `\sigma` of the vertices. For a vertex `v`, we define
    `N_i(v) = \{u | u \in N(v) \text{ and } \sigma(u) < i\}`, that is the subset
    of neighbors of `v` appearing before the `i`-th vertex in the ordering
    `\sigma`. Now, a part of an ordering `\sigma` is a set of consecutive
    vertices, `S = \{u | i \leq \sigma(u) \leq j\}`, such that for any `u \in
    S`, we have `N_i(u) = N_i(\sigma^{-1}(i))` and for any `v` such that `j <
    \sigma(v)`, `N_i(v) \neq N_i(\sigma^{-1}(i))`. The *head* of a part is the
    first position of its vertices.

    The algorithm starts with a single part containing all vertices. Then, when
    the position of the `i`-th vertex `v` is fixed, it explores the neighbors of
    `v` that have not yet been ordered. Consider a part `S` such that `N(x)\cap
    S \neq \emptyset`. The algorithm will rearrange the ordering of the vertices
    in `S` so that the first vertices are the neighbors of `v`. The subpart
    containing the neighbors of `v` is assigned a new name, and the head of `S`
    is set to the position of the first vertex of `S \setminus N(v)` in the
    ordering `\sigma`.

    Observe that each arc of the graph can induce the subdivision of a part.
    Hence, the algorithm can use up to `m + 1` different parts.

    The time complexity of this algorithm is in `O(n + m)`, and our
    implementation follows that complexity ``SparseGraph``. For ``DenseGraph``,
    the complexity is `O(n^2)`. See [HMPV2000]_ and [TCHP2008]_ for more
    details.

    This implementation of extended LexBFS offers some guarantee on the order in
    which the vertices appear in the computed ordering: in case of a tie between
    lexicographic labels during the computation, this function will "choose" the
    vertices in the order in which they appear during the enumeration of the
    neighbors of their last common neighbor.
    For example, if `(u_0, ..., u_k)` is the beginning of the ordering being
    computed and the vertices `v` and `w` currently have the same lexicographic
    label (it means that they have the same neighbors in `(u_0, ..., u_k)`).
    Let call `u_j` their last neighbor in the current ordering (*i.e.*, for all
    `i > j`, `u_i` is not a neighbor of `v` and `w`). This implementation
    will choose `v` for the next vertex of the ordering if and only if `v`
    appeared before `w` when the neighbors of `u_j` where enumerated.

    One possible use of this guarantee is that the caller can reorder the
    adjacency list of vertices (by using, for example, a static sparse graph)
    to force the computed LexBFS order to respect a previous one.

    EXAMPLES::

    To see how it can be used, see the code of the lex_BFS method (in
    traversals.pyx) or of the class SliceDecomposition in this module.

    TESTS:

    Indirect doctests::

        sage: G = graphs.HouseGraph()
        sage: G.slice_decomposition()
        [0[1[2]] [3] [4]]
        sage: G.lex_BFS(algorithm="fast")
        [0, 1, 2, 3, 4]
    """
    cdef int n = <int> cg.num_verts
    # Variables for the partition refinement algorithm
    cdef size_t max_nparts = cg.num_arcs // 2 + 1
    cdef bint need_to_delete_sigma_inv = sigma_inv == NULL
    if sigma_inv == NULL:
        sigma_inv = new vector[int]()
    cdef vector[size_t] part_of = vector[size_t](n, 0)
    cdef vector[size_t] part_len  # only used if xslice_len != NULL (see below)
    cdef vector[size_t] part_head = vector[size_t](max_nparts)
    cdef vector[size_t] subpart = vector[size_t](max_nparts)
    cdef size_t p, part_of_i, nparts, old_nparts
    # Temporary variables
    cdef int max_degree = 0
    cdef int i, j, k, l, u_int, v_int, t_int

    # Resize vectors
    sigma.resize(n)
    deref(sigma_inv).resize(cg.active_vertices.size)
    if pred != NULL:
        deref(pred).clear()
        deref(pred).resize(n, -1)  # initialize pred[i] to -1 for 0 <= i < n
    if xslice_len != NULL:
        deref(xslice_len).resize(n)
        part_len.resize(max_nparts)
    if lex_label != NULL:
        deref(lex_label).resize(n)

    # Initialize the position of vertices in sigma (and compute max_degree)
    if initial_v_int >= 0:
        sigma[0] = initial_v_int
        deref(sigma_inv)[initial_v_int] = 0
        i = 1
    else:
        i = 0
    for v_int in range(<int> cg.active_vertices.size):
        if bitset_in(cg.active_vertices, v_int):
            if v_int != initial_v_int:
                sigma[i] = v_int
                deref(sigma_inv)[v_int] = i
                i = i + 1
            max_degree = max(max_degree, cg.out_degrees[v_int])

    # Variables needed to iterate over neighbors of a vertex
    cdef int nneighbors
    cdef vector[int] neighbors = vector[int](max_degree)

    # Initialize partition: one part containing all the vertices
    nparts = 1
    # all element of part_of are already initialized to 0
    part_head[0] = 0
    subpart[0] = 0
    if xslice_len != NULL:
        part_len[0] = n

    # Main loop
    for i in range(n):
        old_nparts = nparts

        part_of_i = part_of[i]

        # put i out of its part (updating part_len if needed)
        part_head[part_of_i] += 1
        if xslice_len != NULL:
            deref(xslice_len)[i] = part_len[part_of_i]
            part_len[part_of_i] -= 1

        v_int = sigma[i]

        # Iterate over the neighbors of v
        nneighbors = cg.out_neighbors_unsafe (v_int, neighbors.data(), max_degree)
        for k in range(nneighbors):
            u_int = neighbors[k]
            j = deref(sigma_inv)[u_int]  # get the position of u
            if j <= i:
                continue  # already taken care of

            if lex_label != NULL:
                deref(lex_label)[j].push_back (v_int)

            p = part_of[j]  # get the part of u
            l = part_head[p]  # get the beginning of the part containing u

            # if not last and next elem belongs in the same part (ie #part >= 2)
            if l < n - 1 and part_of[l + 1] == p:
                if l != j:  # not already first elem of the part
                    # Place u at the position of the head of the part
                    t_int = sigma[l]
                    deref(sigma_inv)[t_int], deref(sigma_inv)[u_int] = j, l
                    sigma[j], sigma[l] = t_int, u_int
                    if lex_label != NULL:
                        swap[vector[int]](deref(lex_label)[j],
                                          deref(lex_label)[l])
                    j = l
                part_head[p] += 1  # move the head of the part to next elem

            # if part p was not already cut in two during this iteration, we
            # create a new part using subpart
            if subpart[p] < old_nparts:
                subpart[p] = nparts
                part_head[nparts] = j
                if xslice_len != NULL:
                    part_len[nparts] = 0
                subpart[nparts] = 0
                nparts += 1

            # Finally, we update the name of the part for position j and set v
            # as predecessor of u
            part_of[j] = subpart[p]
            if xslice_len != NULL:
                part_len[p] -= 1
                part_len[subpart[p]] += 1
            if pred != NULL:
                deref(pred)[j] = i

    if need_to_delete_sigma_inv:
        del sigma_inv


def slice_decomposition(G, initial_vertex=None):
    r"""
    Compute a slice decomposition of the simple undirected graph

    INPUT:

    - ``G`` -- a Sage graph.

    - ``initial_vertex`` -- (default: ``None``); the first vertex to consider.

    OUTPUT:

    An object of type :class:`~sage.graphs.graph_decompositions.slice_decomposition.SliceDecomposition`
    that represents a slice decomposition of ``G``

    .. NOTE::

        Loops and multiple edges are ignored during the computation of the slice
        decomposition.

    ALGORITHM:

    The method use the algorithm based on "partition refinement" described in
    [HMPV2000]_ and [TCHP2008]_.
    The time complexity of this algorithm is in `O(n + m)`, and our
    implementation follows that complexity for ``SparseGraph``. For
    ``DenseGraph``, the complexity is `O(n^2)`.

    EXAMPLES:

    Slice decomposition of the Petersen Graph::

        sage: G = graphs.PetersenGraph()
        sage: SD = G.slice_decomposition(); SD
        [0[1[4[5]]] [2[6]] [3] [9] [7] [8]]

    The graph can have loops or multiple edges but they are ignored::

        sage: H = Graph(G,loops=True,multiedges=True)
        sage: H.add_edges([(4, 4), (2, 2), (1, 6)])
        sage: SD2 = H.slice_decomposition()
        sage: SD2 == SD
        True
        sage: SD2.underlying_graph() == G.to_simple(immutable=True)
        True

    The tree corresponding to the slice decomposition can be displayed using
    ``view``::

        sage: from sage.graphs.graph_latex import check_tkz_graph
        sage: check_tkz_graph()  # random - depends on Tex installation
        sage: view(G)  # not tested
        sage: latex(G)  # to obtain the corresponding LaTeX code                        # needs sage.plot
        \begin{tikzpicture}
        ...
        \end{tikzpicture}

    Slice decompositions are only defined for undirected graphs::

        sage: from sage.graphs.graph_decompositions.slice_decomposition import slice_decomposition
        sage: slice_decomposition(DiGraph())
        Traceback (most recent call last):
        ...
        ValueError: parameter G must be an undirected graph
    """
    return SliceDecomposition(G, initial_vertex=initial_vertex)


cdef class SliceDecomposition(SageObject):

    def __init__(self, G, initial_vertex=None):
        r"""
        Represents a slice decomposition of a simple directed graph.

        INPUT:

        - ``G`` -- a Sage graph.

        - ``initial_vertex`` -- (default: ``None``); the first vertex to
          consider.

        .. SEEALSO::

            * :meth:`~slice_decomposition` -- compute a slice decomposition of
              the simple undirected graph
            * Section 3.2 of [TCHP2008]_ for a formal definition.

        EXAMPLES:

        The constructor of the :class:`~SliceDecomposition` class is called by
        the :meth:`~slice_decomposition` method of undirected graphs::

            sage: from sage.graphs.graph_decompositions.slice_decomposition import SliceDecomposition
            sage: G = graphs.PetersenGraph()
            sage: SliceDecomposition(G) == G.slice_decomposition()
            True

        The vertex appearing first in the slice decomposition can be specified::

            sage: from sage.graphs.graph_decompositions.slice_decomposition import SliceDecomposition
            sage: SliceDecomposition(graphs.PetersenGraph(), initial_vertex=3)
            [3[2[4[8]]] [1[7]] [0] [9] [6] [5]]

        Slice decompositions are not defined for directed graphs::

            sage: from sage.graphs.graph_decompositions.slice_decomposition import SliceDecomposition
            sage: SliceDecomposition(DiGraph())
            Traceback (most recent call last):
            ...
            ValueError: parameter G must be an undirected graph

        .. automethod:: __getitem__
        """
        if G.is_directed():
            raise ValueError("parameter G must be an undirected graph")

        if initial_vertex is not None and initial_vertex not in G:
            raise LookupError(f"vertex ({initial_vertex}) is not a vertex of the graph")

        cdef CGraphBackend Gbackend = <CGraphBackend> G._backend
        cdef CGraph cg = Gbackend.cg()

        self._graph_class = type(G)

        cdef int initial_v_int
        if initial_vertex is not None:
            # we already checked that initial_vertex is in G
            initial_v_int = Gbackend.get_vertex(initial_vertex)
        else:
            initial_v_int = -1

        cdef vector[int] sigma
        cdef vector[vector[int]] lex_label

        # Compute the slice decomposition using the extended lexBFS algorithm
        extended_lex_BFS(cg, sigma, NULL, initial_v_int, NULL,
                         &(self.xslice_len), &lex_label)

        # Translate the results with the actual vertices of the graph
        self.sigma = tuple(Gbackend.vertex_label(v_int) for v_int in sigma)
        self.sigma_inv = {v: i  for i, v in enumerate(self.sigma)}
        self.lex_label = {i: tuple(Gbackend.vertex_label(v_int) for v_int in lli)
                                    for i, lli in enumerate(lex_label)}

    def __eq__(self, other):
        """
        Return whether ``self`` and ``other`` are equal.

        TESTS::

            sage: G = graphs.PetersenGraph()
            sage: SD = G.slice_decomposition()
            sage: SD == SD
            True
            sage: SD == G.slice_decomposition()
            True

            sage: P3 = graphs.PathGraph(3)
            sage: SD1 = P3.slice_decomposition(initial_vertex=0)
            sage: SD2 = P3.slice_decomposition(initial_vertex=2)
            sage: SD1 == SD2
            False
            sage: SD3 = graphs.CompleteGraph(3).slice_decomposition()
            sage: SD1 == SD3  # same lexBFS but different slice for 1
            False
            sage: SD4 = Graph([(0,1), (0,2)]).slice_decomposition()
            sage: SD3 == SD4  # same lexBFS and slices but different active edges
            False
        """
        if not isinstance(other, type(self)):
            return False

        cdef SliceDecomposition sd = <SliceDecomposition>other

        return self.sigma_inv == sd.sigma_inv \
                and self.lex_label == sd.lex_label \
                and self.xslice_len == sd.xslice_len

    def __hash__(self):
        r"""
        Compute a hash of a ``SliceDecomposition`` object.

        TESTS::

            sage: P3 = graphs.PathGraph(3)
            sage: SD1 = P3.slice_decomposition(initial_vertex=0)
            sage: SD2 = P3.slice_decomposition(initial_vertex=2)
            sage: len({SD1: 1, SD2: 2})  # indirect doctest
            2
        """
        return hash((tuple(self.sigma_inv.items()),
                     tuple(self.lex_label.items()),
                     tuple(self.xslice_len)))

    def __getitem__(self, v):
        r"""
        Return the data about the x-slice of the vertex `v`.

        INPUT:

        - ``v`` -- a vertex of the graph corresponding to the slice
          decomposition.

        OUTPUT:

        A dictionnary with the keys:

        * ``"pivot"`` -- the vertex `v` given as parameter

        * ``"slice"`` -- the slice of `v` (see :meth:`~slice`)

        * ``"active_edges"`` -- the actives edges of `v` (see
          :meth:`~active_edges`)

        * ``"lexicographic_label"`` -- the lexicographic label of `v` (see
          :meth:`~lexicographic_label`)

        * ``"sequence"`` -- the x-slice sequence of `v` (see
          :meth:`~xslice_sequence`)

        This method can also be called via :meth:`xslice_data`.

        EXAMPLES:

        ::

            sage: G = Graph('L~mpn~Nrv{^o~_').relabel('abcdefguvwxyz',inplace=False)
            sage: SD = G.slice_decomposition(initial_vertex='x')
            sage: SD.xslice_data('a')
            {'active_edges': [('a', 'b'),
              ('a', 'c'),
              ('a', 'd'),
              ('a', 'e'),
              ('a', 'f'),
              ('c', 'g'),
              ('d', 'g'),
              ('f', 'g')],
             'lexicographic_label': ['x'],
             'pivot': 'a',
             'sequence': [['a'], ['b', 'c', 'd', 'e', 'f'], ['g']],
             'slice': ['a', 'b', 'c', 'd', 'e', 'f', 'g']}
            sage: SD.xslice_data('u')
            {'active_edges': [],
             'lexicographic_label': ['a', 'b', 'c', 'd', 'e', 'f', 'g'],
             'pivot': 'u',
             'sequence': [['u'], ['y', 'z']],
             'slice': ['u', 'y', 'z']}

        Some values of the returned dictionnary can be obtained via other
        methods (:meth:`~slice`, :meth:`~xslice_sequence`,
        :meth:`~active_edges`, :meth:`~lexicographic_label`)::

            sage: SD.slice('a')
            ['a', 'b', 'c', 'd', 'e', 'f', 'g']
            sage: SD.xslice_data('a')['slice']
            ['a', 'b', 'c', 'd', 'e', 'f', 'g']

            sage: SD.xslice_sequence('a')
            [['a'], ['b', 'c', 'd', 'e', 'f'], ['g']]
            sage: SD.xslice_data('a')['sequence']
            [['a'], ['b', 'c', 'd', 'e', 'f'], ['g']]

            sage: SD.active_edges('b') == SD.xslice_data('b')['active_edges']
            True

            sage: SD.lexicographic_label('u')
            ['a', 'b', 'c', 'd', 'e', 'f', 'g']
            sage: SD.xslice_data('u')['lexicographic_label']
            ['a', 'b', 'c', 'd', 'e', 'f', 'g']

        TESTS::

            sage: G = graphs.RandomGNP(15, 0.3)
            sage: SD = G.slice_decomposition()
            sage: all(SD[v]['slice'] == SD.slice(v) for v in G)
            True
            sage: all(SD[v]['sequence'] == SD.xslice_sequence(v) for v in G)
            True
            sage: all(SD[v]['active_edges'] == SD.active_edges(v) for v in G)
            True
            sage: all(SD[v]['lexicographic_label'] == SD.lexicographic_label(v) for v in G)
            True

            sage: SD = graphs.PetersenGraph().slice_decomposition()
            sage: SD['John']
            Traceback (most recent call last):
            ...
            LookupError: vertex (John) does not appear in the slice decomposition
        """
        if v not in self.sigma_inv:
            raise LookupError(f"vertex ({v}) does not appear in the slice "
                              "decomposition")
        cdef size_t i = self.sigma_inv[v]
        return {'pivot': v,
                'slice': self._slice(i),
                'sequence': self._xslice_sequence(i),
                'lexicographic_label': self._xslice_lex_label(i),
                'active_edges': self._xslice_active_edges(i),
                }

    def lexBFS_order(self):
        r"""
        Return the lexBFS order corresponding to the slice decomposition.

        EXAMPLES::

            sage: from sage.graphs.traversals import _is_valid_lex_BFS_order
            sage: G = graphs.PetersenGraph(); SD = G.slice_decomposition()
            sage: SD.lexBFS_order()
            [0, 1, 4, 5, 2, 6, 3, 9, 7, 8]
            sage: _is_valid_lex_BFS_order(G, SD.lexBFS_order())
            True

        TESTS::

            sage: from sage.graphs.traversals import _is_valid_lex_BFS_order
            sage: for _ in range(5):
            ....:   G = graphs.RandomGNP(15, 0.3)
            ....:   SD = G.slice_decomposition()
            ....:   _is_valid_lex_BFS_order(G, SD.lexBFS_order())
            True
            True
            True
            True
            True
        """
        return list(self.sigma)

    def xslice_data(self, v):
        r"""
        Return the data about the x-slice of the vertex `v`.

        This method is a wrapper around :meth:`SliceDecomposition.__getitem__`

        TESTS::

            sage: G = graphs.RandomGNP(15, 0.3)
            sage: SD = G.slice_decomposition()
            sage: all(SD[v] == SD.xslice_data(v) for v in G)
            True
        """
        return self[v]

    def slice(self, v):
        r"""
        Return the slice of the vertex `v`.

        The slice of `v` is the list of vertices `u` such that the neighbors of
        `u` that are before `v` in the lexBFS order are that same that the
        neighbors of `v` that are before `v` in the lexBFS order (*i.e.*, the
        lexicographic label of `v`). It can be shown that it is a factor of the
        lexBFS order.

        INPUT:

        - ``v`` -- a vertex of the graph corresponding to the slice
          decomposition.

        OUTPUT:

        A list of vertices

        EXAMPLES:

        ::

            sage: G = Graph('L~mpn~Nrv{^o~_').relabel('abcdefguvwxyz',inplace=False)
            sage: SD = G.slice_decomposition(initial_vertex='x')
            sage: SD.slice('a')
            ['a', 'b', 'c', 'd', 'e', 'f', 'g']

        The vertices of the slice have the same neighborhood "on the left"::

            sage: pos = lambda v: SD.lexBFS_order().index(v)
            sage: lla = set(SD.lexicographic_label('a'))
            sage: all(lla == {u for u in G.neighbors(v) if pos(u) < pos('a')} \
            ....:       for v in SD.slice('a'))
            True

        The slice is a factor of the lexBFS order::

            sage: ''.join(SD.slice('a')) in ''.join(SD.lexBFS_order())
            True

        The slice of the initial vertex is the whole graph::

            sage: SD.slice('x') == SD.lexBFS_order()
            True

        TESTS::

            sage: SD.slice('Michael')
            Traceback (most recent call last):
            ...
            LookupError: vertex (Michael) does not appear in the slice decomposition
        """
        if v not in self.sigma_inv:
            raise LookupError(f"vertex ({v}) does not appear in the slice "
                              "decomposition")
        cdef size_t i = self.sigma_inv[v]
        return self._slice(i)

    def xslice_sequence(self, v):
        r"""
        Return the x-slice sequence of the vertex `v`.

        INPUT:

        - ``v`` -- a vertex of the graph corresponding to the slice
          decomposition.

        OUTPUT:

        A list of list corresponding to the x-slice sequence of ``v``.

        EXAMPLES:

        ::

            sage: G = Graph('L~mpn~Nrv{^o~_').relabel('abcdefguvwxyz',inplace=False)
            sage: SD = G.slice_decomposition(initial_vertex='x')
            sage: SD.xslice_sequence('x')
            [['x'], ['a', 'b', 'c', 'd', 'e', 'f', 'g'], ['u', 'y', 'z'], ['v', 'w']]
            sage: SD.xslice_sequence('a')
            [['a'], ['b', 'c', 'd', 'e', 'f'], ['g']]

        The flatten x-slice sequence of a vertex corresponds to the slice of the
        same vertex::

            sage: from itertools import chain
            sage: all(list(chain(*SD.xslice_sequence(v))) == SD.slice(v) \
            ....:       for v in G)
            True

        The first list of the sequence is always a singleton containing the
        input vertex::

            sage: all(SD.xslice_sequence(v)[0] == [v] for v in G)
            True

        If the length of the slice if more than 1, the second list of the
        sequence is either, all the remaining vertices of the slice of `v`, if
        `v` is isolated in the subgraph induced by the slice of `v`, or the
        neighbors of `v` in the subgraph induced by the slice of `v`::

            sage: all(SD.xslice_sequence(v)[1] == SD.slice(v)[1:] for v in G \
            ....:           if G.subgraph(SD.slice(v)).degree(v) == 0 \
            ....:               and len(SD.slice(v)) > 1)
            True
            sage: for v in G:
            ....:     if len(SD.slice(v)) > 1:
            ....:         xslice_seq = SD.xslice_sequence(v)
            ....:         S = G.subgraph(SD.slice(v))
            ....:         if S.degree(v) > 0:
            ....:             set(xslice_seq[1]) == set(S.neighbor_iterator(v))
            True
            True
            True
            True

        TESTS::

            sage: SD = graphs.PetersenGraph().slice_decomposition()
            sage: SD.xslice_sequence('Terry')
            Traceback (most recent call last):
            ...
            LookupError: vertex (Terry) does not appear in the slice decomposition
        """
        if v not in self.sigma_inv:
            raise LookupError(f"vertex ({v}) does not appear in the slice "
                              "decomposition")
        cdef size_t i = self.sigma_inv[v]
        return self._xslice_sequence(i)

    def lexicographic_label(self, v):
        r"""
        Return the lexicographic label of the vertex `v`.

        The lexicographic label of a vertex `v` is the list of all the
        neighbors of `v` that appear before `v` in the lexBFS ordering
        corresponding to the slice decomposition.

        INPUT:

        - ``v`` -- a vertex of the graph corresponding to the slice
          decomposition.

        OUTPUT:

        A list of vertices.

        EXAMPLES::

            sage: G = Graph('L~mpn~Nrv{^o~_').relabel('abcdefguvwxyz',inplace=False)
            sage: SD = G.slice_decomposition(initial_vertex='x')
            sage: SD.lexicographic_label('f')
            ['x', 'a', 'c', 'd']
            sage: pos = lambda v: SD.lexBFS_order().index(v)
            sage: set(SD.lexicographic_label('f')) \
            ....:   == {v for v in G.neighbors('f') if pos(v) < pos('f')}
            True

        TESTS::

            sage: SD = graphs.PetersenGraph().slice_decomposition()
            sage: SD.lexicographic_label('Eric')
            Traceback (most recent call last):
            ...
            LookupError: vertex (Eric) does not appear in the slice decomposition
        """
        if v not in self.sigma_inv:
            raise LookupError(f"vertex ({v}) does not appear in the slice "
                              "decomposition")
        cdef size_t i = self.sigma_inv[v]
        return self._xslice_lex_label(i)

    def active_edges(self, v):
        r"""
        Return the active edges of the vertex `v`.

        An edge `(u, w)` is said to be active for `v` if `u` and `w` belongs
        to two differents slices of the x-slice sequence of `v`. Note that it
        defines a partition of the edges of the underlying graph.

        INPUT:

        - ``v`` -- a vertex of the graph corresponding to the slice
          decomposition.

        OUTPUT:

        A list of edges

        EXAMPLES:

        ::

            sage: G = Graph('L~mpn~Nrv{^o~_').relabel('abcdefguvwxyz',inplace=False)
            sage: SD = G.slice_decomposition(initial_vertex='x')
            sage: SD.xslice_sequence('a')
            [['a'], ['b', 'c', 'd', 'e', 'f'], ['g']]
            sage: ('c', 'g') in SD.active_edges('a')
            True
            sage: ('a', 'c') in SD.active_edges('a')
            True
            sage: ('c', 'd') in SD.active_edges('a')  # c and d in same slice
            False
            sage: ('a', 'u') in SD.active_edges('a')  # u not in x-slice of a
            False

        The set of active edges of every vertex is a partition of the edges::

            sage: from itertools import chain
            sage: E = list(chain(*(SD.active_edges(v) for v in G)))
            sage: G.size() == len(E) == len(set(E)) \
            ....:   and all(G.has_edge(u, w) for v in G for u, w in SD.active_edges(v))
            True

        TESTS::

            sage: SD = graphs.PetersenGraph().slice_decomposition()
            sage: SD.active_edges('Graham')
            Traceback (most recent call last):
            ...
            LookupError: vertex (Graham) does not appear in the slice decomposition
        """
        if v not in self.sigma_inv:
            raise LookupError(f"vertex ({v}) does not appear in the slice "
                              "decomposition")
        cdef size_t i = self.sigma_inv[v]
        return self._xslice_active_edges(i)

    def _slice(self, size_t idx):
        r"""
        This method is for internal use only

        TESTS:

        Indirect doctests::

            sage: SD = graphs.HouseGraph().slice_decomposition()
            sage: SD.slice(1)
            [1, 2]
        """
        return list(self.sigma[idx:idx+self.xslice_len[idx]])

    def _xslice_sequence(self, size_t idx):
        r"""
        This method is for internal use only

        TESTS:

        Indirect doctests::

            sage: SD = graphs.HouseGraph().slice_decomposition()
            sage: SD.xslice_sequence(0)
            [[0], [1, 2], [3], [4]]
        """
        cdef size_t l = self.xslice_len[idx]
        cdef size_t j = idx + 1
        cdef size_t lj

        S = [ [self.sigma[idx]] ]
        while j < idx + l:
            lj = self.xslice_len[j]
            S.append(list(self.sigma[j:j+lj]))
            j += lj
        assert j == idx + l, "slice decomposition is ill-formed"
        return S

    def _xslice_lex_label(self, size_t idx):
        r"""
        This method is for internal use only

        TESTS:

        Indirect doctests::

            sage: SD = graphs.HouseGraph().slice_decomposition()
            sage: SD.lexicographic_label(3)
            [1, 2]
        """
        return list(self.lex_label[idx])

    def _xslice_active_edges(self, size_t idx):
        r"""
        This method is for internal use only

        TESTS:

        Indirect doctests::

            sage: SD = graphs.HouseGraph().slice_decomposition()
            sage: SD.active_edges(0)
            [(0, 1), (0, 2), (1, 3), (2, 3), (2, 4), (3, 4)]
        """
        cdef size_t l = self.xslice_len[idx]
        cdef size_t llv_prefix = len(self.lex_label[idx])
        cdef size_t j = idx + 1
        cdef size_t lj

        A = []
        while j < idx + l:
            lj = self.xslice_len[j]
            llj = self.lex_label[j]
            for u in self.sigma[j:j+lj]:
                for w in llj[llv_prefix:]:
                    A.append((w, u))
            j += lj
        assert j == idx + l, "slice decomposition is ill-formed"
        return A

    def underlying_graph(self):
        r"""
        Return the underlying graph corresponding to the slice decomposition.

        If `G` was the graph given as parameter to compute the slice
        decomposition, the underlying graph corresponds to ``G.to_simple()``
        where labels are ignored, *i.e.*, it is the input graph without loops,
        multiple edges and labels.

        .. NOTE::

            This method is mostly defined to test the computation of
            lexicographic labels and actives edges.

        EXAMPLES:

        ::

            sage: G = Graph('L~mpn~Nrv{^o~_').relabel('abcdefguvwxyz',inplace=False)
            sage: SD = G.slice_decomposition(initial_vertex='x')
            sage: SD.underlying_graph() == G
            True

        The graph can have loops or multiple edges but they are ignored::

            sage: G = graphs.CubeConnectedCycle(2)  # multiple edges
            sage: SD = G.slice_decomposition()
            sage: SD.underlying_graph() == G.to_simple(immutable=True)
            True

            sage: G = graphs.CubeConnectedCycle(1)  # loops
            sage: SD = G.slice_decomposition()
            sage: SD.underlying_graph() == G.to_simple(immutable=True)
            True

        TESTS::

            sage: for _ in range(5):
            ....:   G = graphs.RandomGNP(15, 0.3)
            ....:   SD = G.slice_decomposition()
            ....:   SD.underlying_graph() == G
            True
            True
            True
            True
            True
        """
        if not hasattr(self, '_underlying_graph'):
            vertices = self.sigma
            edges = [(u, v) for i, v in enumerate(self.sigma)
                            for u in self.lex_label[i]]
            data = [vertices, edges]
            Gclass = self._graph_class
            self._underlying_graph = Gclass(data, format='vertices_and_edges',
                                            immutable=True)
        return self._underlying_graph

    def _repr_(self):
        r"""
        Return a string representation of a ``SliceDecomposition`` object.

        TESTS::

            sage: G = graphs.PetersenGraph(); SD = G.slice_decomposition()
            sage: repr(SD)
            '[0[1[4[5]]] [2[6]] [3] [9] [7] [8]]'
            sage: G = Graph('L~mpn~Nrv{^o~_').relabel('abcdefguvwxyz',inplace=False)
            sage: SD = G.slice_decomposition(initial_vertex='x'); repr(SD)
            '[x[a[b[c[d]] [e[f]]] [g]] [u[y[z]]] [v[w]]]'
        """
        def inner_repr(idx):
            l = self.xslice_len[idx]
            S = []
            if l > 1:
                j = idx + 1
                while j < idx + l:
                    lj = self.xslice_len[j]
                    S.append(inner_repr(j))
                    j += lj
                assert j == idx + l, "slice decomposition is ill-formed"
            return f'{self.sigma[idx]}' + ' '.join(f'[{s}]' for s in S)
        return f'[{inner_repr(0)}]'

    def _latex_(self):
        r"""
        Return a string to render, using `\LaTeX`, the slice decomposition as a
        tree.

        TESTS::

            sage: from sage.graphs.graph_latex import check_tkz_graph
            sage: check_tkz_graph()  # random - depends on Tex installation
            sage: G = graphs.PetersenGraph(); SD = G.slice_decomposition()
            sage: latex(SD)
            \begin{tikzpicture}
            ...
              v0 -- {l0, v1, v4, v6, v7, v8, v9};
              v1 -- {l1, v2};
              v2 -- {l2, v3};
              v3 -- {l3};
              v4 -- {l4, v5};
              v5 -- {l5};
              v6 -- {l6};
              v7 -- {l7};
              v8 -- {l8};
              v9 -- {l9};
            ...
            \end{tikzpicture}

        """
        from sage.misc.latex import latex

        latex.add_package_to_preamble_if_available("tikz")
        latex.add_to_preamble(r"\usetikzlibrary{arrows,shapes,fit}")
        latex.add_to_preamble(r"\usetikzlibrary{graphs,graphdrawing}")
        latex.add_to_preamble(r"\usegdlibrary{trees}")

        # Call latex() on all vertices
        sigma_latex = [ latex(v) for v in self.sigma ]
        slices = [[] for _ in self.sigma]

        lines = [ r"\begin{tikzpicture}" ]
        lines.append(r"\graph [tree layout,level distance=0,level sep=1em,"
                     r"sibling distance=0,sibling sep=0.6em,"
                     r"tail anchor=center,head anchor=north,"
                     r"nodes={draw,rectangle,inner xsep=0.2em},edges={thick}]")
        lines.append("{")
        bo, bc = "{", "}"  # to write { and } in f-strings
        # Create the nodes and leaves of the slice decomposition tree
        for i in range(len(self.sigma)):
            l = self.xslice_len[i]
            label = r"\ ".join(sigma_latex[i:i+l])
            lines.append(f"  v{i}[as={bo}${label}${bc}];")
            lines.append(f"  l{i}[draw=none,as={bo}${sigma_latex[i]}${bc}];")
            j = i + 1
            slices[i].append(f"l{i}")
            while j < i + l:
                slices[i].append(f"v{j}")
                j += self.xslice_len[j]
        # Create the edges of the slice decomposition tree
        for i, S in enumerate(slices):
            lines.append(f"  v{i} -- " + "{" + ", ".join(S) + "};")
        lines.append("};")
        # Add dahsed red boxes around xslices
        for i, S in enumerate(slices):
            fit=" ".join(f"({s})" for s in S)
            lines.append(rf"\node (s{i}) [rectangle,inner xsep=0.2em,draw=red,"
                         f"densely dashed,fit={fit}]{bo}{bc};")

        lines.append(r"\end{tikzpicture}")
        return "\n".join(lines)



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.libs.gap
r"""
Database of generalised quadrangles with spread

This module implements some construction of generalised quadrangles
with spread.

EXAMPLES::

    sage: GQ, S = designs.generalised_quadrangle_with_spread(4, 16, check=False)
    sage: GQ
    Incidence structure with 325 points and 1105 blocks
    sage: GQ2, O = designs.generalised_quadrangle_hermitian_with_ovoid(4)
    sage: GQ2
    Incidence structure with 1105 points and 325 blocks
    sage: GQ3 = GQ.dual()
    sage: set(GQ3._points) == set(GQ2._points)
    True
    sage: GQ2.is_isomorphic(GQ3) # long time
    True

REFERENCES:

- [PT2009]_

- [TP1994]_

- :wikipedia:`Generalized_quadrangle`

AUTHORS:

- Ivo Maffei (2020-07-26): initial version
"""

# ****************************************************************************
#       Copyright (C) 2020 Ivo Maffei <ivomaffei@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


def generalised_quadrangle_with_spread(const int s, const int t,
                                       existence=False, check=True):
    r"""
    Construct a generalised quadrangle GQ of order `(s,t)` with a spread S.

    INPUT:

    - ``s``, ``t`` -- integers; order of the generalised quadrangle

    - ``existence`` -- boolean;

    - ``check`` -- boolean; if ``True``, then Sage checks that the object built
      is correct. (default: ``True``)

    OUTPUT:

    A pair `(GQ, S)` where `GQ` is a :class:`IncidenceStructure` representing
    the generalised quadrangle and `S` is a list of blocks of `GQ` representing
    the spread of `GQ`.

    EXAMPLES::

        sage: t = designs.generalised_quadrangle_with_spread(3, 9)
        sage: t[0]
        Incidence structure with 112 points and 280 blocks
        sage: designs.generalised_quadrangle_with_spread(5, 25, existence=True)
        True
        sage: (designs.generalised_quadrangle_with_spread(4, 16, check=False))[0]
        Incidence structure with 325 points and 1105 blocks
        sage: designs.generalised_quadrangle_with_spread(0, 2, existence=True)
        False

    REFERENCES:

    For more on generalised quadrangles and their spread see [PT2009]_ or
    [TP1994]_.

    TESTS::

        sage: GQ, S = designs.generalised_quadrangle_with_spread(2, 4)
        sage: GQ
        Incidence structure with 27 points and 45 blocks
        sage: designs.generalised_quadrangle_with_spread(3, 4)
        Traceback (most recent call last):
        ...
        RuntimeError: Sage can't build a GQ of order (3, 4) with a spread
        sage: designs.generalised_quadrangle_with_spread(3, 4, existence=True)
        Unknown
    """
    from sage.combinat.designs.incidence_structures import IncidenceStructure
    from sage.misc.unknown import Unknown
    from sage.arith.misc import is_prime_power

    if s < 1 or t < 1:
        if existence:
            return False
        raise RuntimeError(f"No GQ of order ({s}, {t}) exists")

    if s == 1 and t == 1:  # we have a square
        if existence:
            return True
        D = IncidenceStructure([[0, 1], [1, 2], [2, 3], [3, 0]])
        return (D, [[0, 1], [2, 3]])

    if is_prime_power(s) and t == s * s:
        if existence:
            return True
        (GQ, S) = dual_GQ_ovoid(*generalised_quadrangle_hermitian_with_ovoid(s))
        if check:
            if not is_GQ_with_spread(GQ, S, s=s, t=t):
                raise RuntimeError("Sage built a wrong GQ with spread")
        return (GQ, S)

    if existence:
        return Unknown
    raise RuntimeError(
        f"Sage can't build a GQ of order ({s}, {t}) with a spread")


def is_GQ_with_spread(GQ, S, s=None, t=None):
    r"""
    Check if GQ is a generalised quadrangle of order `(s,t)` and
    check that S is a spread of GQ

    INPUT:

    - ``GQ`` -- IncidenceStructure; the incidence structure that is supposed to
      be a generalised quadrangle

    - ``S`` -- iterable; the spread of ``GQ`` as an
      iterable of the blocks of ``GQ``

    - ``s``, ``t`` -- integers (optional); if `(s,t)` are given, then we check that
      ``GQ`` has order `(s,t)`

    EXAMPLES::

        sage: from sage.combinat.designs.gen_quadrangles_with_spread import *
        sage: t = generalised_quadrangle_hermitian_with_ovoid(3)
        sage: is_GQ_with_spread(*t)
        Traceback (most recent call last):
        ...
        TypeError: 'int' object is not iterable
        sage: t = dual_GQ_ovoid(*t)
        sage: is_GQ_with_spread(*t)
        True
        sage: is_GQ_with_spread(*t, s=3)
        True

    TESTS::

       sage: from sage.combinat.designs.gen_quadrangles_with_spread import *
       sage: t = generalised_quadrangle_hermitian_with_ovoid(2)
       sage: t = dual_GQ_ovoid(*t)
       sage: is_GQ_with_spread(*t, s=2, t=4)
       True
       sage: is_GQ_with_spread(*t, s=2)
       True
       sage: is_GQ_with_spread(*t, s=3)
       False
    """
    res = GQ.is_generalized_quadrangle(parameters=True)
    if res is False \
       or (s is not None and s != res[0]) \
       or (t is not None and t != res[1]):
        return False

    # check spread
    return GQ.is_spread(S)


def dual_GQ_ovoid(GQ, O):
    r"""
    Compute the dual incidence structure of GQ
    and return the image of `O` under the dual map

    INPUT:

    - ``GQ`` -- IncidenceStructure; the generalised quadrangle we want
      the dual of

    - ``O`` -- iterable; the iterable of blocks we want to compute the dual

    OUTPUT: a pair ``(D, S)`` where ``D`` is the dual of ``GQ`` and
    ``S`` is the dual of ``O``

    EXAMPLES::

        sage: from sage.combinat.designs.gen_quadrangles_with_spread import \
        ....: dual_GQ_ovoid
        sage: t = designs.generalised_quadrangle_hermitian_with_ovoid(3)
        sage: t[0].is_generalized_quadrangle(parameters=True)
        (9, 3)
        sage: t = dual_GQ_ovoid(*t)
        sage: t[0].is_generalized_quadrangle(parameters=True)
        (3, 9)
        sage: all([x in t[0] for x in t[1]])
        True


    TESTS::

        sage: from sage.combinat.designs.gen_quadrangles_with_spread import *
        sage: t = designs.generalised_quadrangle_hermitian_with_ovoid(2)
        sage: t = dual_GQ_ovoid(*t)
        sage: t[0].is_generalized_quadrangle(parameters=True)
        (2, 4)
        sage: is_GQ_with_spread(*t)
        True
    """
    from sage.combinat.designs.incidence_structures import IncidenceStructure

    # GQ.ground_set()[i] becomes newBlocks[i]
    # GQ.blocks()[i] becomes i
    newBlocks = [[] for _ in range(GQ.num_points())]
    pointsToInt = {p: i for i, p in enumerate(GQ.ground_set())}

    for i, b in enumerate(GQ.blocks()):
        for p in b:
            newBlocks[pointsToInt[p]].append(i)

    S = [newBlocks[pointsToInt[p]] for p in O]

    D = IncidenceStructure(newBlocks)
    return (D, S)


def generalised_quadrangle_hermitian_with_ovoid(const int q):
    r"""
    Construct the generalised quadrangle `H(3,q^2)` with an ovoid.

    The GQ has order `(q^2,q)`.

    INPUT:

    - ``q`` -- integer; a prime power

    OUTPUT:

    A pair ``(D, O)`` where ``D`` is an IncidenceStructure representing the
    generalised quadrangle and ``O`` is a list of points of ``D`` which
    constitute an ovoid of ``D``.

    EXAMPLES::

        sage: t = designs.generalised_quadrangle_hermitian_with_ovoid(4)
        sage: t[0]
        Incidence structure with 1105 points and 325 blocks
        sage: len(t[1])
        65
        sage: G = t[0].intersection_graph([1])  # line graph
        sage: G.is_strongly_regular(True)
        (325, 68, 3, 17)
        sage: set(t[0].block_sizes())
        {17}

    REFERENCES:

    For more on `H(3,q^2)` and the construction implemented here see [PT2009]_
    or [TP1994]_.

    TESTS::

        sage: from sage.combinat.designs.gen_quadrangles_with_spread import \
        ....: is_GQ_with_spread, dual_GQ_ovoid
        sage: t = designs.generalised_quadrangle_hermitian_with_ovoid(3)
        sage: t = dual_GQ_ovoid(*t)
        sage: is_GQ_with_spread(*t, s=3, t=9)
        True
        sage: t = dual_GQ_ovoid(*(
        ....: designs.generalised_quadrangle_hermitian_with_ovoid(2)))
        sage: t[0]
        Incidence structure with 27 points and 45 blocks
        sage: len(t[1])
        9
    """
    from sage.libs.gap.libgap import libgap
    from sage.combinat.designs.incidence_structures import IncidenceStructure
    from sage.arith.misc import is_prime_power

    GU = libgap.GU(4, q)
    Fq = libgap.GF(q * q)
    zero = libgap.Zero(Fq)
    one = libgap.One(Fq)
    V = libgap.FullRowSpace(Fq, 4)

    e1 = [one, zero, zero, zero]  # isotropic point

    points = list(libgap.Orbit(GU, e1, libgap.OnLines))  # all isotropic points
    pointInt = { x: int(i + 1) for i, x in enumerate(points) }
    # above we sum 1 because GAP starts at 1

    GUp = libgap.Action(GU, points, libgap.OnLines)

    e2 = [zero, one, zero, zero]  # another isotropic point
    line = V.Subspace([e1, e2])  # totally isotropic line
    lineAsPoints = [libgap.Elements(libgap.Basis(b))[0]
                    for b in libgap.Elements(line.Subspaces(1))]
    line = libgap.Set([pointInt[p] for p in lineAsPoints])

    lines = libgap.Orbit(GUp, line, libgap.OnSets)  # all isotropic lines
    lines = [list(map(lambda x: int(x - 1), b)) for b in lines]  # convert to int
    # lines defines the GQ H(3, q^2)

    # to find an ovoid, we embed H(3,q^2) in H(4,q^2)
    # the embedding is (a,b,c,d) -> (a,b,0,c,d)
    # then we find a point in the latter and not in the former
    # this point will be collinear (in H(3,q^2)) to all points in an ovoid
    if q % 2 == 1:
        p, _ = is_prime_power(q, get_data=True)
        a = (p - 1) // 2
        aGap = zero
        for i in range(a):
            aGap += one
        p = [zero, one, one, aGap, zero]
    else:
        a = libgap.PrimitiveRoot(Fq)**(q-1)
        p = [zero, one, a+one, a, zero]

    J = [[0, 0, 0, 0, 1], [0, 0, 0, 1, 0], [0, 0, 1, 0, 0],
         [0, 1, 0, 0, 0], [1, 0, 0, 0, 0]]
    J = libgap(J)  # matrix of the invariant form of GU(5,q)

    # p' is collinear to p iff p'Jp^q = 0
    # note that p'Jp^q = bx^q + c where p' = (a,b,0,c,d) and p = (0,1,y,x,0)
    ovoid = []
    xq = p[3]**q
    for p2 in points:
        if p2[1]*xq + p2[2] == zero:  # p collinear to p2
            ovoid.append(pointInt[p2] - 1)

    D = IncidenceStructure(lines)
    return (D, ovoid)



# sage_setup: distribution = sagemath-graphs
r"""
Paths in directed acyclic graphs
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.structure.parent import Parent
import sage.graphs.digraph as digraph


def GraphPaths(g, source=None, target=None):
    """
    Return the combinatorial class of paths in the directed acyclic graph g.

    EXAMPLES::

        sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)

    If source and target are not given, then the returned class
    contains all paths (including trivial paths containing only one
    vertex).

    ::

        sage: p = GraphPaths(G); p
        Paths in Multi-digraph on 5 vertices
        sage: p.cardinality()
        37
        sage: path = p.random_element()
        sage: all(G.has_edge(*path[i:i+2]) for i in range(len(path) -1))
        True

    If the source is specified, then the returned class contains all of
    the paths starting at the vertex source (including the trivial
    path).

    ::

        sage: p = GraphPaths(G, source=3); p
        Paths in Multi-digraph on 5 vertices starting at 3
        sage: p.list()
        [[3], [3, 4], [3, 4, 5], [3, 4, 5]]

    If the target is specified, then the returned class contains all of
    the paths ending at the vertex target (including the trivial
    path).

    ::

        sage: p = GraphPaths(G, target=3); p
        Paths in Multi-digraph on 5 vertices ending at 3
        sage: p.cardinality()
        5
        sage: p.list()
        [[3], [1, 3], [2, 3], [1, 2, 3], [1, 2, 3]]

    If both the target and source are specified, then the returned
    class contains all of the paths from source to target.

    ::

        sage: p = GraphPaths(G, source=1, target=3); p
        Paths in Multi-digraph on 5 vertices starting at 1 and ending at 3
        sage: p.cardinality()
        3
        sage: p.list()
        [[1, 2, 3], [1, 2, 3], [1, 3]]

    Note that G must be a directed acyclic graph.

    ::

        sage: G = DiGraph({1:[2,2,3,5], 2:[3,4], 3:[4], 4:[2,5,7], 5:[6]}, multiedges=True)
        sage: GraphPaths(G)
        Traceback (most recent call last):
        ...
        TypeError: g must be a directed acyclic graph
    """
    if not isinstance(g, digraph.DiGraph):
        raise TypeError("g must be a DiGraph")
    elif not g.is_directed_acyclic():
        raise TypeError("g must be a directed acyclic graph")

    if source is None and target is None:
        return GraphPaths_all(g)
    elif source is not None and target is None:
        if source not in g:
            raise ValueError("source must be in g")
        return GraphPaths_s(g, source)
    elif source is None and target is not None:
        if target not in g:
            raise ValueError("target must be in g")
        return GraphPaths_t(g, target)
    else:
        if source not in g:
            raise ValueError("source must be in g")
        if target not in g:
            raise ValueError("target must be in g")
        return GraphPaths_st(g, source, target)


class GraphPaths_common:
    def __eq__(self, other):
        """
        Test for equality.

        EXAMPLES::

            sage: G1 = DiGraph({1:[2,3], 2:[3]})
            sage: p1 = GraphPaths(G1)
            sage: G2 = DiGraph({2:[3], 3:[4]})
            sage: p2 = GraphPaths(G2)
            sage: p1 == p1
            True
            sage: p1 == p2
            False
        """
        if not isinstance(other, GraphPaths_common):
            return False
        return self.graph == other.graph

    def __ne__(self, other):
        """
        Test for unequality.

        EXAMPLES::

            sage: G1 = DiGraph({1:[2,3], 2:[3]})
            sage: p1 = GraphPaths(G1)
            sage: G2 = DiGraph({2:[3], 3:[4]})
            sage: p2 = GraphPaths(G2)
            sage: p1 != p2
            True
            sage: p1 != p1
            False
        """
        return not (self == other)

    def outgoing_edges(self, v):
        """
        Return a list of v's outgoing edges.

        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G)
            sage: p.outgoing_edges(2)
            [(2, 3, None), (2, 4, None)]
        """
        return list(self.graph.outgoing_edge_iterator(v))

    def incoming_edges(self, v):
        """
        Return a list of v's incoming edges.

        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G)
            sage: p.incoming_edges(2)
            [(1, 2, None), (1, 2, None)]
        """
        return list(self.graph.incoming_edge_iterator(v))

    def outgoing_paths(self, v):
        """
        Return a list of the paths that start at v.

        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: gp = GraphPaths(G)
            sage: gp.outgoing_paths(3)
            [[3], [3, 4], [3, 4, 5], [3, 4, 5]]
            sage: gp.outgoing_paths(2)
            [[2],
             [2, 3],
             [2, 3, 4],
             [2, 3, 4, 5],
             [2, 3, 4, 5],
             [2, 4],
             [2, 4, 5],
             [2, 4, 5]]
        """
        source_paths = [[v]]
        for e in self.outgoing_edges(v):
            target = e[1]
            target_paths = self.outgoing_paths(target)
            target_paths = [[v] + path for path in target_paths]

            source_paths += target_paths

        return source_paths

    def incoming_paths(self, v):
        """
        Return a list of paths that end at v.

        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: gp = GraphPaths(G)
            sage: gp.incoming_paths(2)
            [[2], [1, 2], [1, 2]]
        """
        target_paths = [[v]]
        for e in self.incoming_edges(v):
            source = e[0]
            source_paths = self.incoming_paths(source)
            source_paths = [path + [v] for path in source_paths]
            target_paths += source_paths
        return target_paths

    def paths_from_source_to_target(self, source, target):
        """
        Return a list of paths from source to target.

        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: gp = GraphPaths(G)
            sage: gp.paths_from_source_to_target(2,4)
            [[2, 3, 4], [2, 4]]
        """
        source_paths = self.outgoing_paths(source)
        return [path for path in source_paths if path[-1] == target]

    def paths(self):
        """
        Return a list of all the paths of ``self``.

        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: gp = GraphPaths(G)
            sage: len(gp.paths())
            37
        """
        paths = []
        for source in self.graph.vertex_iterator():
            paths += self.outgoing_paths(source)
        return paths


class GraphPaths_all(Parent, GraphPaths_common):
    """
    EXAMPLES::

        sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
        sage: p = GraphPaths(G)
        sage: p.cardinality()
        37
    """

    def __init__(self, g):
        """
        TESTS::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G)
            sage: p == loads(dumps(p))
            True
        """
        self.graph = g
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def __repr__(self):
        """
        TESTS::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G)
            sage: repr(p)
            'Paths in Multi-digraph on 5 vertices'
        """
        return "Paths in %s" % repr(self.graph)

    def list(self):
        """
        Return a list of the paths of ``self``.

        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: len(GraphPaths(G).list())
            37
        """
        return self.paths()


class GraphPaths_t(Parent, GraphPaths_common):
    def __init__(self, g, target):
        """
        TESTS::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G, target=4)
            sage: p == loads(dumps(p))
            True
        """
        self.graph = g
        self.target = target
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def __repr__(self):
        """
        TESTS::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G, target=4)
            sage: repr(p)
            'Paths in Multi-digraph on 5 vertices ending at 4'
        """
        return "Paths in %s ending at %s" % (repr(self.graph), self.target)

    def list(self):
        """
        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G, target=4)
            sage: p.list()
            [[4],
             [2, 4],
             [1, 2, 4],
             [1, 2, 4],
             [3, 4],
             [1, 3, 4],
             [2, 3, 4],
             [1, 2, 3, 4],
             [1, 2, 3, 4]]
        """
        return self.incoming_paths(self.target)


class GraphPaths_s(Parent, GraphPaths_common):
    def __init__(self, g, source):
        """
        TESTS::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G, 4)
            sage: p == loads(dumps(p))
            True
        """
        self.graph = g
        self.source = source
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def __repr__(self):
        """
        TESTS::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G, 4)
            sage: repr(p)
            'Paths in Multi-digraph on 5 vertices starting at 4'
        """
        return "Paths in %s starting at %s" % (repr(self.graph), self.source)

    def list(self):
        """
        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G, 4)
            sage: p.list()
            [[4], [4, 5], [4, 5]]
        """
        return self.outgoing_paths(self.source)


class GraphPaths_st(Parent, GraphPaths_common):
    """
    EXAMPLES::

        sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
        sage: GraphPaths(G,1,2).cardinality()
        2
        sage: GraphPaths(G,1,3).cardinality()
        3
        sage: GraphPaths(G,1,4).cardinality()
        5
        sage: GraphPaths(G,1,5).cardinality()
        10
        sage: GraphPaths(G,2,3).cardinality()
        1
        sage: GraphPaths(G,2,4).cardinality()
        2
        sage: GraphPaths(G,2,5).cardinality()
        4
        sage: GraphPaths(G,3,4).cardinality()
        1
        sage: GraphPaths(G,3,5).cardinality()
        2
        sage: GraphPaths(G,4,5).cardinality()
        2
    """

    def __init__(self, g, source, target):
        """
        TESTS::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G,1,2)
            sage: p == loads(dumps(p))
            True
        """
        self.graph = g
        self.source = source
        self.target = target
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def __repr__(self):
        """
        TESTS::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G,1,2)
            sage: repr(p)
            'Paths in Multi-digraph on 5 vertices starting at 1 and ending at 2'
        """
        return "Paths in %s starting at %s and ending at %s" % (repr(self.graph), self.source, self.target)

    def list(self):
        """
        EXAMPLES::

            sage: G = DiGraph({1:[2,2,3], 2:[3,4], 3:[4], 4:[5,5]}, multiedges=True)
            sage: p = GraphPaths(G,1,2)
            sage: p.list()
            [[1, 2], [1, 2]]
        """
        return self.paths_from_source_to_target(self.source, self.target)
